from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from datetime import datetime
from typing import TYPE_CHECKING, Generic, Optional, cast

from typing_extensions import TypeVar

from dagster._annotations import public
from dagster._core.definitions.partitions.context import (
    PartitionLoadingContext,
    partition_loading_context,
)
from dagster._core.definitions.partitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.partitions.utils.base import (
    generate_partition_key_based_definition_id,
)
from dagster._core.errors import DagsterInvalidInvocationError, DagsterUnknownPartitionError
from dagster._core.storage.tags import PARTITION_NAME_TAG
from dagster._core.types.pagination import PaginatedResults

if TYPE_CHECKING:
    from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
    from dagster._core.instance import DynamicPartitionsStore


T_str = TypeVar("T_str", bound=str, default=str, covariant=True)


@public
class PartitionsDefinition(ABC, Generic[T_str]):
    """Defines a set of partitions, which can be attached to a software-defined asset or job.

    Abstract class with implementations for different kinds of partitions.
    """

    @property
    def partitions_subset_class(self) -> type["PartitionsSubset"]:
        from dagster._core.definitions.partitions.subset.default import DefaultPartitionsSubset

        return DefaultPartitionsSubset

    def validate_partition_definition(self) -> None:
        pass

    @abstractmethod
    @public
    def get_partition_keys(
        self,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> Sequence[T_str]:
        """Returns a list of strings representing the partition keys of the PartitionsDefinition.

        Args:
            current_time (Optional[datetime]): A datetime object representing the current time, only
                applicable to time-based partitions definitions.
            dynamic_partitions_store (Optional[DynamicPartitionsStore]): The DynamicPartitionsStore
                object that is responsible for fetching dynamic partitions. Required when the
                partitions definition is a DynamicPartitionsDefinition with a name defined. Users
                can pass the DagsterInstance fetched via `context.instance` to this argument.

        Returns:
            Sequence[str]
        """
        ...

    @abstractmethod
    def get_paginated_partition_keys(
        self,
        context: PartitionLoadingContext,
        limit: int,
        ascending: bool,
        cursor: str | None = None,
    ) -> PaginatedResults[str]:
        """Returns a connection object that contains a list of partition keys and all the necessary
        information to paginate through them.

        Args:
            context (PartitionLoadingContext): The context for loading partition keys.
            limit (int): The maximum number of partition keys to return.
            ascending (bool): Whether to return the partition keys in ascending order.  The order is determined by the partitions definition.
            cursor (Optional[str]): A cursor to track the progress paginating through the returned partition key results.

        Returns:
            PaginatedResults[str]
        """
        ...

    def __str__(self) -> str:
        joined_keys = ", ".join([f"'{key}'" for key in self.get_partition_keys()])
        return joined_keys

    def get_last_partition_key(self) -> T_str | None:
        partition_keys = self.get_partition_keys()
        return partition_keys[-1] if partition_keys else None

    def get_first_partition_key(self) -> T_str | None:
        partition_keys = self.get_partition_keys()
        return partition_keys[0] if partition_keys else None

    def get_partition_keys_in_range(
        self, partition_key_range: PartitionKeyRange
    ) -> Sequence[T_str]:
        keys_exist = {
            partition_key_range.start: self.has_partition_key(partition_key_range.start),
            partition_key_range.end: self.has_partition_key(partition_key_range.end),
        }
        if not all(keys_exist.values()):
            raise DagsterInvalidInvocationError(
                f"""Partition range {partition_key_range.start} to {partition_key_range.end} is
                not a valid range. Nonexistent partition keys:
                {list(key for key in keys_exist if keys_exist[key] is False)}"""
            )

        # in the simple case, simply return the single key in the range
        if partition_key_range.start == partition_key_range.end:
            return [cast("T_str", partition_key_range.start)]

        # defer this call as it is potentially expensive
        partition_keys = self.get_partition_keys()
        return partition_keys[
            partition_keys.index(partition_key_range.start) : partition_keys.index(
                partition_key_range.end
            )
            + 1
        ]

    def empty_subset(self) -> "PartitionsSubset":
        return self.partitions_subset_class.create_empty_subset(self)

    def subset_with_partition_keys(self, partition_keys: Iterable[str]) -> "PartitionsSubset":
        return self.empty_subset().with_partition_keys(partition_keys)

    def subset_with_all_partitions(self) -> "PartitionsSubset":
        return self.subset_with_partition_keys(self.get_partition_keys())

    def deserialize_subset(self, serialized: str) -> "PartitionsSubset":
        return self.partitions_subset_class.from_serialized(self, serialized)

    def can_deserialize_subset(
        self,
        serialized: str,
        serialized_partitions_def_unique_id: str | None,
        serialized_partitions_def_class_name: str | None,
    ) -> bool:
        return self.partitions_subset_class.can_deserialize(
            self,
            serialized,
            serialized_partitions_def_unique_id,
            serialized_partitions_def_class_name,
        )

    def get_serializable_unique_identifier(self) -> str:
        return generate_partition_key_based_definition_id(self.get_partition_keys())

    def get_tags_for_partition_key(self, partition_key: str) -> Mapping[str, str]:
        tags = {PARTITION_NAME_TAG: partition_key}
        return tags

    def get_num_partitions(self) -> int:
        return len(self.get_partition_keys())

    def has_partition_key(self, partition_key: str) -> bool:
        return partition_key in self.get_partition_keys()

    def validate_partition_key(self, partition_key: str, context: PartitionLoadingContext) -> None:
        with partition_loading_context(new_ctx=context):
            if not self.has_partition_key(partition_key):
                raise DagsterUnknownPartitionError(
                    f"Could not find a partition with key `{partition_key}`."
                )



from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, AbstractSet, Annotated, Any  # noqa: UP035

from dagster._check import checked
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
from dagster._core.definitions.events import AssetKey
from dagster._core.errors import DagsterUserCodeProcessError
from dagster._core.execution.plan.state import KnownExecutionState
from dagster._core.instance import DagsterInstance
from dagster._core.remote_origin import RemoteJobOrigin
from dagster._core.remote_representation.external_data import DEFAULT_MODE_NAME
from dagster._core.snap.execution_plan_snapshot import (
    ExecutionPlanSnapshot,
    ExecutionPlanSnapshotErrorData,
)
from dagster._grpc.types import ExecutionPlanSnapshotArgs
from dagster._record import ImportFrom
from dagster._serdes import deserialize_value

if TYPE_CHECKING:
    from dagster._grpc.client import DagsterGrpcClient


@checked
def sync_get_external_execution_plan_grpc(
    api_client: Annotated["DagsterGrpcClient", ImportFrom("dagster._grpc.client")],
    job_origin: RemoteJobOrigin,
    run_config: Mapping[str, Any],
    job_snapshot_id: str,
    asset_selection: AbstractSet[AssetKey] | None = None,
    asset_check_selection: AbstractSet[AssetCheckKey] | None = None,
    op_selection: Sequence[str] | None = None,
    step_keys_to_execute: Sequence[str] | None = None,
    known_state: KnownExecutionState | None = None,
    instance: DagsterInstance | None = None,
) -> ExecutionPlanSnapshot:
    op_selection = op_selection or []
    result = deserialize_value(
        api_client.execution_plan_snapshot(
            execution_plan_snapshot_args=ExecutionPlanSnapshotArgs(
                job_origin=job_origin,
                op_selection=op_selection,
                run_config=run_config,
                mode=DEFAULT_MODE_NAME,
                step_keys_to_execute=step_keys_to_execute,
                job_snapshot_id=job_snapshot_id,
                known_state=known_state,
                instance_ref=instance.get_ref() if instance and instance.is_persistent else None,
                asset_selection=asset_selection,
                asset_check_selection=asset_check_selection,
            )
        ),
        (ExecutionPlanSnapshot, ExecutionPlanSnapshotErrorData),
    )

    if isinstance(result, ExecutionPlanSnapshotErrorData):
        raise DagsterUserCodeProcessError.from_error_info(result.error)
    return result


@checked
async def gen_external_execution_plan_grpc(
    api_client: Annotated["DagsterGrpcClient", ImportFrom("dagster._grpc.client")],
    job_origin: RemoteJobOrigin,
    run_config: Mapping[str, Any],
    job_snapshot_id: str,
    asset_selection: AbstractSet[AssetKey] | None = None,
    asset_check_selection: AbstractSet[AssetCheckKey] | None = None,
    op_selection: Sequence[str] | None = None,
    step_keys_to_execute: Sequence[str] | None = None,
    known_state: KnownExecutionState | None = None,
    instance: DagsterInstance | None = None,
) -> ExecutionPlanSnapshot:
    op_selection = op_selection or []

    result = deserialize_value(
        await api_client.gen_execution_plan_snapshot(
            execution_plan_snapshot_args=ExecutionPlanSnapshotArgs(
                job_origin=job_origin,
                op_selection=op_selection,
                run_config=run_config,
                mode=DEFAULT_MODE_NAME,
                step_keys_to_execute=step_keys_to_execute,
                job_snapshot_id=job_snapshot_id,
                known_state=known_state,
                instance_ref=instance.get_ref() if instance and instance.is_persistent else None,
                asset_selection=asset_selection,
                asset_check_selection=asset_check_selection,
            )
        ),
        (ExecutionPlanSnapshot, ExecutionPlanSnapshotErrorData),
    )

    if isinstance(result, ExecutionPlanSnapshotErrorData):
        raise DagsterUserCodeProcessError.from_error_info(result.error)
    return result



from collections.abc import Sequence
from datetime import datetime
from typing import TYPE_CHECKING, NamedTuple, Optional

from dagster._annotations import PublicAttr
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition import PartitionsDefinition
from dagster._core.definitions.partitions.mapping.partition_mapping import (
    PartitionMapping,
    UpstreamPartitionsResult,
)
from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
from dagster._serdes import whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.instance import DynamicPartitionsStore


@whitelist_for_serdes
class SpecificPartitionsPartitionMapping(
    PartitionMapping,
    NamedTuple(
        "_SpecificPartitionsPartitionMapping", [("partition_keys", PublicAttr[Sequence[str]])]
    ),
):
    """Maps to a specific subset of partitions in the upstream asset.

    Example:
        .. code-block:: python

            from dagster import SpecificPartitionsPartitionMapping, StaticPartitionsDefinition, asset

            @asset(partitions_def=StaticPartitionsDefinition(["a", "b", "c"]))
            def upstream():
                ...

            @asset(
                ins={
                    "upstream": AssetIn(partition_mapping=SpecificPartitionsPartitionMapping(["a"]))
                }
            )
            def a_downstream(upstream):
                ...
    """

    def validate_partition_mapping(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition | None,
    ):
        pass

    def get_upstream_mapped_partitions_result_for_partitions(
        self,
        downstream_partitions_subset: PartitionsSubset | None,
        downstream_partitions_def: PartitionsDefinition | None,
        upstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> UpstreamPartitionsResult:
        with partition_loading_context(current_time, dynamic_partitions_store):
            return UpstreamPartitionsResult(
                partitions_subset=upstream_partitions_def.subset_with_partition_keys(
                    self.partition_keys
                ),
                required_but_nonexistent_subset=upstream_partitions_def.empty_subset(),
            )

    def get_downstream_partitions_for_partitions(
        self,
        upstream_partitions_subset: PartitionsSubset,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> PartitionsSubset:
        with partition_loading_context(current_time, dynamic_partitions_store):
            # if any of the partition keys in this partition mapping are contained within the upstream
            # partitions subset, then all partitions of the downstream asset are dependencies
            if any(key in upstream_partitions_subset for key in self.partition_keys):
                return downstream_partitions_def.subset_with_all_partitions()
            return downstream_partitions_def.empty_subset()

    @property
    def description(self) -> str:
        return f"Each downstream partition depends on the following upstream partitions: {self.partition_keys}"



from typing import NamedTuple

from dagster_shared.serdes import deserialize_value
from dagster_shared.serdes.errors import DeserializationError

import dagster._check as check
from dagster._serdes import whitelist_for_serdes


@whitelist_for_serdes
class AssetDetails(NamedTuple("_AssetDetails", [("last_wipe_timestamp", float | None)])):
    """Set of asset fields that do not change with every materialization.  These are generally updated
    on some non-materialization action (e.g. wipe).
    """

    def __new__(cls, last_wipe_timestamp: float | None = None):
        check.opt_float_param(last_wipe_timestamp, "last_wipe_timestamp")
        return super().__new__(cls, last_wipe_timestamp)

    @staticmethod
    def from_db_string(db_string):
        if not db_string:
            return None

        try:
            details = deserialize_value(db_string, AssetDetails)
        except DeserializationError:
            return None

        return details



from collections.abc import Iterable
from typing import TYPE_CHECKING, Optional

from dagster_shared import record
from dagster_shared.serdes import whitelist_for_serdes

import dagster._check as check
from dagster._core.definitions.asset_health.asset_health import AssetHealthStatus
from dagster._core.definitions.asset_health.asset_materialization_health import (
    MinimalAssetMaterializationHealthState,
)
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.freshness import FreshnessState, FreshnessStateRecord
from dagster._core.loader import LoadableBy, LoadingContext
from dagster._core.storage.event_log.base import AssetRecord

if TYPE_CHECKING:
    from dagster._core.workspace.context import BaseWorkspaceRequestContext


@whitelist_for_serdes
@record.record
class AssetFreshnessHealthState(LoadableBy[AssetKey]):
    """Maintains the latest freshness state for the asset."""

    freshness_state: FreshnessState
    updated_timestamp: float | None = None

    @property
    def health_status(self) -> AssetHealthStatus:
        if self.freshness_state == FreshnessState.PASS:
            return AssetHealthStatus.HEALTHY
        elif self.freshness_state == FreshnessState.WARN:
            return AssetHealthStatus.WARNING
        elif self.freshness_state == FreshnessState.FAIL:
            return AssetHealthStatus.DEGRADED
        elif self.freshness_state == FreshnessState.NOT_APPLICABLE:
            return AssetHealthStatus.NOT_APPLICABLE
        else:
            return AssetHealthStatus.UNKNOWN

    @classmethod
    async def compute_for_asset(
        cls, asset_key: AssetKey, loading_context: LoadingContext
    ) -> "AssetFreshnessHealthState":
        """Gets the freshness state for the asset from the DB."""
        freshness_state_record = await FreshnessStateRecord.gen(loading_context, asset_key)

        if freshness_state_record is None:
            # freshness policy has no evaluations yet
            return cls(
                freshness_state=FreshnessState.UNKNOWN,
                updated_timestamp=None,
            )
        return cls(
            freshness_state=freshness_state_record.freshness_state,
            updated_timestamp=freshness_state_record.updated_at.timestamp(),
        )

    @classmethod
    def _blocking_batch_load(
        cls, keys: Iterable[AssetKey], context: LoadingContext
    ) -> Iterable[Optional["AssetFreshnessHealthState"]]:
        asset_freshness_health_states = (
            context.instance.get_asset_freshness_health_state_for_assets(list(keys))
        )
        return [asset_freshness_health_states.get(key) for key in keys]


@whitelist_for_serdes
@record.record
class AssetHealthFreshnessMetadata:
    last_materialized_timestamp: float | None


async def get_freshness_status_and_metadata(
    context: "BaseWorkspaceRequestContext", asset_key: AssetKey
) -> tuple[AssetHealthStatus, Optional["AssetHealthFreshnessMetadata"]]:
    """Gets an AssetFreshnessHealthState object for an asset, either via streamline or by computing
    it based on the state of the DB. Then converts it to a AssetHealthStatus and the metadata
    needed to power the UIs. Metadata is computed based on the state of the DB.
    """
    asset_freshness_health_state = await AssetFreshnessHealthState.gen(context, asset_key)
    if (
        asset_freshness_health_state is None
    ):  # if streamline reads are off or no streamline state exists for the asset compute it from the DB
        if context.instance.streamline_read_asset_health_required("asset-freshness-health"):
            return AssetHealthStatus.UNKNOWN, None

        if (
            not context.asset_graph.has(asset_key)
            or context.asset_graph.get(asset_key).freshness_policy is None
        ):
            return AssetHealthStatus.NOT_APPLICABLE, None
        asset_freshness_health_state = await AssetFreshnessHealthState.compute_for_asset(
            asset_key,
            context,
        )

    asset_materialization_health_state = await MinimalAssetMaterializationHealthState.gen(
        context, asset_key
    )
    if asset_materialization_health_state is None:
        asset_record = await AssetRecord.gen(context, asset_key)
        materialization_timestamp = (
            asset_record.asset_entry.last_materialization.timestamp
            if asset_record
            and asset_record.asset_entry
            and asset_record.asset_entry.last_materialization
            else None
        )
    else:
        materialization_timestamp = (
            asset_materialization_health_state.latest_materialization_timestamp
        )

    if asset_freshness_health_state.freshness_state == FreshnessState.PASS:
        return AssetHealthStatus.HEALTHY, AssetHealthFreshnessMetadata(
            last_materialized_timestamp=materialization_timestamp,
        )
    if asset_freshness_health_state.freshness_state == FreshnessState.WARN:
        return AssetHealthStatus.WARNING, AssetHealthFreshnessMetadata(
            last_materialized_timestamp=materialization_timestamp,
        )
    if asset_freshness_health_state.freshness_state == FreshnessState.FAIL:
        return AssetHealthStatus.DEGRADED, AssetHealthFreshnessMetadata(
            last_materialized_timestamp=materialization_timestamp,
        )
    elif asset_freshness_health_state.freshness_state == FreshnessState.UNKNOWN:
        return AssetHealthStatus.UNKNOWN, None
    elif asset_freshness_health_state.freshness_state == FreshnessState.NOT_APPLICABLE:
        return AssetHealthStatus.NOT_APPLICABLE, None

    else:
        check.failed(f"Unexpected freshness state: {asset_freshness_health_state.freshness_state}")



import inspect
from typing import Any, TypeAlias, TypeVar

from dagster_shared import check

# For the time being, `Annotatable` is set to `Any` even though it should be set to `Decoratable` to
# avoid choking the type checker. Choking happens because of a niche scenario where
# `ResourceDefinition`, which is part of `Decoratable`, is used as an argument to a function that
# accepts `Annotatable`. There is a certain circularity here that current versions of pyright fail
# on. It will likely be resolved by future versions of pyright, and then `Annotatable` should be set
# to `Decoratable`.
Annotatable: TypeAlias = Any

T_Annotatable = TypeVar("T_Annotatable", bound=Annotatable)


def _get_annotation_target(obj: Annotatable) -> object:
    """Given an object to be annotated, return the underlying object that will actually store the annotations.
    This is necessary because not all objects are mutable, and so can't be annotated directly.
    """
    if isinstance(obj, property):
        return obj.fget
    elif isinstance(obj, (staticmethod, classmethod)):
        return obj.__func__
    else:
        return obj


def _get_subject(obj: Annotatable, param: str | None = None) -> str:
    from dagster._core.decorator_utils import is_resource_def

    """Get the string representation of an annotated object that will appear in
    annotation-generated warnings about the object.
    """
    if param:
        if isinstance(obj, type):
            return f"Parameter `{param}` of initializer `{obj.__qualname__}.__init__`"
        else:
            fn_subject = _get_subject(obj)
            return f"Parameter `{param}` of {fn_subject[:1].lower() + fn_subject[1:]}"
    else:
        if isinstance(obj, type):
            return f"Class `{obj.__qualname__}`"
        elif isinstance(obj, property):
            return f"Property `{obj.fget.__qualname__ if obj.fget else obj}`"
        # classmethod and staticmethod don't themselves get a `__qualname__` attr until Python 3.10.
        elif isinstance(obj, classmethod):
            return f"Class method `{_get_annotation_target(obj).__qualname__}`"
        elif isinstance(obj, staticmethod):
            return f"Static method `{_get_annotation_target(obj).__qualname__}`"
        elif inspect.isfunction(obj):
            return f"Function `{obj.__qualname__}`"
        elif is_resource_def(obj):
            return f"Dagster resource `{obj.__qualname__}`"
        else:
            check.failed(f"Unexpected object type: {type(obj)}")


def copy_annotations(dest: Annotatable, src: Annotatable) -> None:
    """Copy all Dagster annotations from one object to another object."""
    dest_target = _get_annotation_target(dest)
    src_target = _get_annotation_target(src)

    # Copy all known annotation attributes
    annotation_attrs = [
        "_is_public",
        "_deprecated",
        "_deprecated_params",
        "_superseded",
        "_preview",
        "_beta",
        "_beta_params",
    ]

    for attr_name in annotation_attrs:
        if hasattr(src_target, attr_name):
            setattr(dest_target, attr_name, getattr(src_target, attr_name))



"""Base revision -- noop.

Revision ID: da7cd32b690d
Revises: 567bc23fd1ac
Create Date: 2019-11-22 11:02:06.180581

"""

# revision identifiers, used by Alembic.
revision = "da7cd32b690d"
down_revision = "567bc23fd1ac"
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



"""Factory functions for creating DagsterInstance instances."""

import os
from collections.abc import Sequence
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._core.errors import DagsterHomeNotSetError, DagsterInvariantViolationError
from dagster._core.instance.config import DAGSTER_CONFIG_YAML_FILENAME
from dagster._core.instance.ref import InstanceRef
from dagster._core.instance.types import InstanceType

if TYPE_CHECKING:
    from dagster._core.debug import DebugRunPayload
    from dagster._core.instance.instance import DagsterInstance, DagsterInstanceOverrides


def create_ephemeral_instance(
    tempdir: str | None = None,
    preload: Sequence["DebugRunPayload"] | None = None,
    settings: dict | None = None,
) -> "DagsterInstance":
    """Create a `DagsterInstance` suitable for ephemeral execution, useful in test contexts. An
    ephemeral instance uses mostly in-memory components. Use `local_temp` to create a test
    instance that is fully persistent.

    Args:
        tempdir (Optional[str]): The path of a directory to be used for local artifact storage.
        preload (Optional[Sequence[DebugRunPayload]]): A sequence of payloads to load into the
            instance's run storage. Useful for debugging.
        settings (Optional[Dict]): Settings for the instance.

    Returns:
        DagsterInstance: An ephemeral DagsterInstance.
    """
    from dagster._core.instance.instance import DagsterInstance
    from dagster._core.launcher.sync_in_memory_run_launcher import SyncInMemoryRunLauncher
    from dagster._core.run_coordinator import DefaultRunCoordinator
    from dagster._core.storage.event_log import InMemoryEventLogStorage
    from dagster._core.storage.noop_compute_log_manager import NoOpComputeLogManager
    from dagster._core.storage.root import LocalArtifactStorage, TemporaryLocalArtifactStorage
    from dagster._core.storage.runs import InMemoryRunStorage

    if tempdir is not None:
        local_storage = LocalArtifactStorage(tempdir)
    else:
        local_storage = TemporaryLocalArtifactStorage()

    return DagsterInstance(
        instance_type=InstanceType.EPHEMERAL,
        local_artifact_storage=local_storage,
        run_storage=InMemoryRunStorage(preload=preload),
        event_storage=InMemoryEventLogStorage(preload=preload),
        compute_log_manager=NoOpComputeLogManager(),
        run_coordinator=DefaultRunCoordinator(),
        run_launcher=SyncInMemoryRunLauncher(),
        settings=settings,
    )


def create_instance_from_dagster_home() -> "DagsterInstance":
    """Get the current `DagsterInstance` as specified by the ``DAGSTER_HOME`` environment variable.

    Returns:
        DagsterInstance: The current DagsterInstance.
    """
    dagster_home_path = os.getenv("DAGSTER_HOME")

    if not dagster_home_path:
        raise DagsterHomeNotSetError(
            "The environment variable $DAGSTER_HOME is not set. \nDagster requires this"
            " environment variable to be set to an existing directory in your filesystem. This"
            " directory is used to store metadata across sessions, or load the dagster.yaml"
            " file which can configure storing metadata in an external database.\nYou can"
            " resolve this error by exporting the environment variable. For example, you can"
            " run the following command in your shell or include it in your shell configuration"
            ' file:\n\texport DAGSTER_HOME=~"/dagster_home"\nor PowerShell\n$env:DAGSTER_HOME'
            " = ($home + '\\dagster_home')or batchset"
            " DAGSTER_HOME=%UserProfile%/dagster_homeAlternatively, DagsterInstance.ephemeral()"
            " can be used for a transient instance.\n"
        )

    dagster_home_path = os.path.expanduser(dagster_home_path)

    if not os.path.isabs(dagster_home_path):
        raise DagsterInvariantViolationError(
            f'$DAGSTER_HOME "{dagster_home_path}" must be an absolute path. Dagster requires this '
            "environment variable to be set to an existing directory in your filesystem."
        )

    if not (os.path.exists(dagster_home_path) and os.path.isdir(dagster_home_path)):
        raise DagsterInvariantViolationError(
            f'$DAGSTER_HOME "{dagster_home_path}" is not a directory or does not exist. Dagster requires this'
            " environment variable to be set to an existing directory in your filesystem"
        )

    return create_instance_from_config(dagster_home_path)


def create_local_temp_instance(
    tempdir: str | None = None,
    overrides: Optional["DagsterInstanceOverrides"] = None,
) -> "DagsterInstance":
    """Create a DagsterInstance that uses a temporary directory for local storage. This is a
    regular, fully persistent instance. Use `ephemeral` to get an ephemeral instance with
    in-memory components.

    Args:
        tempdir (Optional[str]): The path of a directory to be used for local artifact storage.
        overrides (Optional[DagsterInstanceOverrides]): Override settings for the instance.

    Returns:
        DagsterInstance
    """
    from dagster._core.instance.instance import DagsterInstance

    if tempdir is None:
        created_dir = TemporaryDirectory()
        i = create_instance_from_ref(InstanceRef.from_dir(created_dir.name, overrides=overrides))
        # Store the temp dir for cleanup when the instance is garbage collected
        DagsterInstance._TEMP_DIRS[i] = created_dir  # noqa: SLF001
        return i

    return create_instance_from_ref(InstanceRef.from_dir(tempdir, overrides=overrides))


def create_instance_from_config(
    config_dir: str,
    config_filename: str = DAGSTER_CONFIG_YAML_FILENAME,
) -> "DagsterInstance":
    instance_ref = InstanceRef.from_dir(config_dir, config_filename=config_filename)
    return create_instance_from_ref(instance_ref)


def create_instance_from_ref(instance_ref: InstanceRef) -> "DagsterInstance":
    from dagster._core.instance.instance import DagsterInstance

    check.inst_param(instance_ref, "instance_ref", InstanceRef)

    # DagsterInstance doesn't implement ConfigurableClass, but we may still sometimes want to
    # have custom subclasses of DagsterInstance. This machinery allows for those custom
    # subclasses to receive additional keyword arguments passed through the config YAML.
    klass = instance_ref.custom_instance_class or DagsterInstance
    kwargs = instance_ref.custom_instance_class_config

    unified_storage = instance_ref.storage
    run_storage = unified_storage.run_storage if unified_storage else instance_ref.run_storage
    event_storage = (
        unified_storage.event_log_storage if unified_storage else instance_ref.event_storage
    )
    schedule_storage = (
        unified_storage.schedule_storage if unified_storage else instance_ref.schedule_storage
    )

    return klass(
        instance_type=InstanceType.PERSISTENT,
        local_artifact_storage=instance_ref.local_artifact_storage,
        run_storage=run_storage,  # type: ignore  # (possible none)
        event_storage=event_storage,  # type: ignore  # (possible none)
        schedule_storage=schedule_storage,
        compute_log_manager=None,  # lazy load
        scheduler=instance_ref.scheduler,
        run_coordinator=None,  # lazy load
        run_launcher=None,  # lazy load
        settings=instance_ref.settings,
        secrets_loader=instance_ref.secrets_loader,
        defs_state_storage=instance_ref.defs_state_storage,
        ref=instance_ref,
        **kwargs,
    )



import abc
import inspect
from collections.abc import Callable, Iterator, Mapping, Sequence
from contextlib import nullcontext
from functools import cached_property
from pathlib import Path
from typing import Generic, TypeVar

from dagster_shared.record import record
from dagster_shared.serdes.objects import EnvRegistryKey
from dagster_shared.seven import load_module_object
from dagster_shared.yaml_utils import parse_yamls_with_source_position
from dagster_shared.yaml_utils.source_position import SourcePosition, ValueAndSourcePositionTree
from pydantic import BaseModel, TypeAdapter

import dagster._check as check
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._utils.pydantic_yaml import (
    _parse_and_populate_model_with_annotated_errors,
    enrich_validation_errors_with_source_position,
)
from dagster.components.component.component import Component, EmptyAttributesModel
from dagster.components.component.component_loader import is_component_loader
from dagster.components.core.context import ComponentDeclLoadContext, ComponentLoadContext
from dagster.components.core.defs_module import (
    EXPLICITLY_IGNORED_GLOB_PATTERNS,
    ComponentFileModel,
    ComponentPath,
    CompositeYamlComponent,
    DefsFolderComponent,
    PythonFileComponent,
    asset_post_processor_list_from_post_processing_dict,
    context_with_injected_scope,
    find_defs_or_component_yaml,
)
from dagster.components.resolved.core_models import AssetPostProcessor

T = TypeVar("T", bound=Component)


@record
class ComponentDecl(abc.ABC, Generic[T]):
    """Class representing a not-yet-loaded component in the defs hierarchy. In effect, a closure
    with all information needed to load the component. A tree of component decls is initially
    built before any components are loaded.
    """

    context: ComponentDeclLoadContext
    path: ComponentPath

    @abc.abstractmethod
    def _load_component(self) -> T:
        """Loads the component represented by this decl."""
        ...

    @property
    @abc.abstractmethod
    def component_type(self) -> type[T]: ...

    def iterate_all_component_decls(self) -> Iterator["ComponentDecl"]:
        for _, component in self.iterate_path_component_decl_pairs():
            yield component

    def iterate_child_component_decls(self) -> Iterator["ComponentDecl"]:
        return iter([])

    def iterate_path_component_decl_pairs(
        self,
    ) -> Iterator[tuple[ComponentPath, "ComponentDecl"]]:
        yield self.path, self
        for child in self.iterate_child_component_decls():
            yield from child.iterate_path_component_decl_pairs()


@record
class ComponentLoaderDecl(ComponentDecl[Component]):
    """Declaration of a component that is loaded by a user-defined Python function."""

    component_node_fn: Callable[[ComponentDeclLoadContext], Component]

    def _load_component(self) -> Component:
        return self.component_node_fn(self.context)

    @property
    def component_type(self) -> type[Component]:
        # parse from function return type if possible
        sig = inspect.signature(self.component_node_fn)
        return sig.return_annotation or Component


@record
class PythonFileDecl(ComponentDecl[PythonFileComponent]):
    """Declaration of a PythonFileComponent, corresponding to a Python file with zero or more
    ComponentLoaderDecls and zero or more plain Dagster defs.
    """

    decls: Mapping[str, ComponentLoaderDecl]

    def _load_component(self) -> "PythonFileComponent":
        return PythonFileComponent(
            components={
                attr: self.context.load_structural_component_at_path(decl.path)
                for attr, decl in self.decls.items()
            },
            path=self.path.file_path,
        )

    @property
    def component_type(self) -> type[PythonFileComponent]:
        return PythonFileComponent

    def iterate_child_component_decls(self) -> Iterator["ComponentDecl"]:
        yield from self.decls.values()


def _get_component_class(
    context: ComponentDeclLoadContext, component_file_model: ComponentFileModel
) -> type:
    # TODO: lookup in cache so we don't have to import the class directly
    type_str = context.normalize_component_type_str(component_file_model.type)
    key = EnvRegistryKey.from_typename(type_str)
    obj = load_module_object(key.namespace, key.name)
    if not isinstance(obj, type) or not issubclass(obj, Component):
        raise DagsterInvalidDefinitionError(
            f"Component type {type_str} is of type {type(obj)}, but must be a subclass of dagster.Component"
        )

    return obj


def _process_attributes_with_enriched_validation_err(
    source_tree: ValueAndSourcePositionTree | None,
    component_file_model: ComponentFileModel | None,
    model_cls: type[BaseModel] | None,
) -> BaseModel | None:
    if (
        model_cls is None
        or not source_tree
        or not component_file_model
        or not component_file_model.attributes
    ):
        return None

    attributes_position_tree = source_tree.source_position_tree.children.get("attributes", None)
    with (
        enrich_validation_errors_with_source_position(attributes_position_tree, ["attributes"])
        if attributes_position_tree
        else nullcontext()
    ):
        if issubclass(model_cls, EmptyAttributesModel):
            raise DagsterInvalidDefinitionError(
                "Component is not resolvable from YAML, but attributes were provided. This error can be avoided by "
                "making the component a subclass of `Resolvable` or by overriding `get_model_cls` on your component."
            )

        return TypeAdapter(model_cls).validate_python(component_file_model.attributes)


@record
class YamlBackedComponentDecl(ComponentDecl[T]):
    source_tree: ValueAndSourcePositionTree | None
    component_file_model: ComponentFileModel | None

    @cached_property
    def context_with_component_injected_scope(self) -> ComponentDeclLoadContext:
        if not self.component_file_model:
            return self.context
        return context_with_injected_scope(
            self.context,
            self.component_type,
            self.component_file_model.template_vars_module,
        )

    @property
    def component_type(self) -> type[T]:
        """The class of the component that is being loaded."""
        return _get_component_class(self.context, check.not_none(self.component_file_model))

    def get_asset_post_processor_lists(self) -> list[AssetPostProcessor]:
        if not self.source_tree or not self.component_file_model:
            return []

        post_processing_position_tree = self.source_tree.source_position_tree.children.get(
            "post_processing", None
        )
        with (
            enrich_validation_errors_with_source_position(
                post_processing_position_tree, ["post_processing"]
            )
            if post_processing_position_tree
            else nullcontext()
        ):
            return asset_post_processor_list_from_post_processing_dict(
                self.context_with_component_injected_scope.resolution_context,
                self.component_file_model.post_processing,
            )


@record
class YamlDecl(YamlBackedComponentDecl):
    """Declaration of a single component loaded from a YAML file."""

    @staticmethod
    def from_source_tree(
        context: ComponentDeclLoadContext,
        source_tree: ValueAndSourcePositionTree,
        path: ComponentPath,
    ) -> "YamlDecl":
        component_file_model = _parse_and_populate_model_with_annotated_errors(
            cls=ComponentFileModel, obj_parse_root=source_tree, obj_key_path_prefix=[]
        )
        return YamlDecl(
            context=context,
            source_tree=source_tree,
            component_file_model=component_file_model,
            path=path,
        )

    def _load_component(self) -> "Component":
        context = self.context_with_component_injected_scope.with_source_position_tree(
            check.not_none(self.source_tree).source_position_tree,
        )

        model_cls = self.component_type.get_model_cls()

        attributes = _process_attributes_with_enriched_validation_err(
            self.source_tree, self.component_file_model, model_cls
        )

        return self.component_type.load(
            attributes, ComponentLoadContext.from_decl_load_context(context, self)
        )


@record
class YamlFileDecl(ComponentDecl[CompositeYamlComponent]):
    """Declaration of a CompositeYamlComponent, corresponding to a YAML file with one or more
    YamlDecls.
    """

    decls: Sequence[YamlBackedComponentDecl]
    source_positions: Sequence[SourcePosition]

    def _load_component(self) -> "CompositeYamlComponent":
        return CompositeYamlComponent(
            components=[
                self.context.load_structural_component_at_path(decl.path) for decl in self.decls
            ],
            source_positions=self.source_positions,
            asset_post_processor_lists=[
                decl.get_asset_post_processor_lists() for decl in self.decls
            ],
        )

    @property
    def component_type(self) -> type[CompositeYamlComponent]:
        return CompositeYamlComponent

    def iterate_child_component_decls(self) -> Iterator["ComponentDecl"]:
        yield from self.decls


@record
class DefsFolderDecl(YamlBackedComponentDecl[DefsFolderComponent]):
    children: Mapping[Path, ComponentDecl]

    @classmethod
    def get(cls, context: ComponentDeclLoadContext) -> "DefsFolderDecl":
        component = build_component_decl_from_context(context)
        return check.inst(
            component,
            DefsFolderDecl,
            f"Expected DefsFolderDecl at {context.path}, got {component}.",
        )

    @property
    def component_type(self) -> type[DefsFolderComponent]:
        return DefsFolderComponent

    def _load_component(self) -> "DefsFolderComponent":
        _process_attributes_with_enriched_validation_err(
            self.source_tree, self.component_file_model, DefsFolderComponent.get_model_cls()
        )
        return DefsFolderComponent(
            path=self.path.file_path,
            children={
                subpath: self.context.load_structural_component_at_path(decl.path)
                for subpath, decl in self.children.items()
            },
        )

    def iterate_child_component_decls(self) -> Iterator["ComponentDecl"]:
        yield from self.children.values()


def build_component_decl_from_context(context: ComponentDeclLoadContext) -> ComponentDecl | None:
    """Attempts to determine the type of component that should be loaded for the given context.  Iterates through potential component
    type matches, prioritizing more specific types: YAML, Python, plain Dagster defs, and component
    folder.
    """
    # in priority order
    # yaml component
    if find_defs_or_component_yaml(context.path):
        return build_component_decl_from_yaml_file_backcompat(context)
    # defs
    elif (
        context.terminate_autoloading_on_keyword_files
        and (context.path / "definitions.py").exists()
        and not (context.path / "component.py").exists()
    ):
        return PythonFileDecl(
            context=context,
            path=ComponentPath.from_path(context.path / "definitions.py"),
            decls={},
        )
    elif context.path.suffix == ".py":
        return build_component_decl_from_python_file(context)
    # folder
    elif context.path.is_dir():
        children = build_component_decls_from_directory_items(context, None)
        if children:
            return DefsFolderDecl(
                context=context,
                path=ComponentPath.from_path(context.path),
                children=children,
                source_tree=None,
                component_file_model=None,
            )

    return None


def build_component_decls_from_directory_items(
    context: ComponentDeclLoadContext, component_file_model: ComponentFileModel | None
) -> Mapping[Path, ComponentDecl]:
    found = {}
    for subpath in sorted(context.path.iterdir()):
        relative_subpath = subpath.relative_to(context.path)
        if any(relative_subpath.match(pattern) for pattern in EXPLICITLY_IGNORED_GLOB_PATTERNS):
            continue
        path_context = context_with_injected_scope(
            context,
            DefsFolderComponent,
            component_file_model.template_vars_module if component_file_model else None,
        ).for_component_path(ComponentPath.from_path(subpath))

        component_node = build_component_decl_from_context(path_context)
        if component_node:
            found[subpath] = component_node
    return found


def build_component_decl_from_python_file(
    context: ComponentDeclLoadContext,
) -> ComponentLoaderDecl | PythonFileDecl:
    # backcompat for component.yaml
    component_def_path = context.path
    module = context.load_defs_relative_python_module(component_def_path)
    component_loaders = list(inspect.getmembers(module, is_component_loader))

    return PythonFileDecl(
        path=ComponentPath.from_path(context.path),
        context=context,
        decls={
            attr: ComponentLoaderDecl(
                context=context.for_component_path(ComponentPath.from_path(context.path, attr)),
                component_node_fn=component_loader,
                path=ComponentPath.from_path(context.path, attr),
            )
            for attr, component_loader in component_loaders
        },
    )


def build_component_decl_from_yaml_file_backcompat(
    context: ComponentDeclLoadContext,
) -> ComponentDecl:
    component_def_path = check.not_none(find_defs_or_component_yaml(context.path))
    return build_component_decl_from_yaml_file(
        context=context, component_def_path=component_def_path
    )


def build_component_decl_from_yaml_file(
    context: ComponentDeclLoadContext, component_def_path: Path
) -> ComponentDecl:
    source_trees = parse_yamls_with_source_position(
        component_def_path.read_text(), str(component_def_path)
    )
    component_nodes = []
    for i, source_tree in enumerate(source_trees):
        component_nodes.append(
            build_component_decl_from_yaml_document(
                context=context.for_component_path(ComponentPath.from_path(context.path, i)),
                source_tree=source_tree,
                path=ComponentPath.from_path(context.path, i),
            )
        )

    check.invariant(len(component_nodes) > 0, "No components found in YAML file")
    return YamlFileDecl(
        path=ComponentPath.from_path(context.path),
        context=context,
        decls=component_nodes,
        source_positions=[
            source_tree.source_position_tree.position for source_tree in source_trees
        ],
    )


def build_component_decl_from_yaml_document(
    context: ComponentDeclLoadContext,
    source_tree: ValueAndSourcePositionTree,
    path: ComponentPath,
) -> ComponentDecl:
    component_file_model = _parse_and_populate_model_with_annotated_errors(
        cls=ComponentFileModel, obj_parse_root=source_tree, obj_key_path_prefix=[]
    )
    component_cls = _get_component_class(context, component_file_model)
    component_decl_type = component_cls.get_decl_type()
    check.invariant(component_decl_type in (YamlDecl, DefsFolderDecl))

    if component_decl_type == DefsFolderDecl:
        children = build_component_decls_from_directory_items(context, component_file_model)
        return DefsFolderDecl(
            context=context,
            path=path,
            children=children,
            source_tree=source_tree,
            component_file_model=component_file_model,
        )
    else:
        return YamlDecl(
            context=context,
            source_tree=source_tree,
            component_file_model=component_file_model,
            path=path,
        )



from enum import Enum

import dagster._check as check
from dagster._serdes import whitelist_for_serdes


@whitelist_for_serdes
class ScheduleType(Enum):
    HOURLY = "HOURLY"
    DAILY = "DAILY"
    WEEKLY = "WEEKLY"
    MONTHLY = "MONTHLY"

    @property
    def ordinal(self):
        return {"HOURLY": 1, "DAILY": 2, "WEEKLY": 3, "MONTHLY": 4}[self.value]

    def __gt__(self, other: "ScheduleType") -> bool:
        check.inst_param(
            other, "other", ScheduleType, "Cannot compare ScheduleType with non-ScheduleType"
        )
        return self.ordinal > other.ordinal

    def __lt__(self, other: "ScheduleType") -> bool:
        check.inst_param(
            other, "other", ScheduleType, "Cannot compare ScheduleType with non-ScheduleType"
        )
        return self.ordinal < other.ordinal


def cron_schedule_from_schedule_type_and_offsets(
    schedule_type: ScheduleType,
    minute_offset: int,
    hour_offset: int,
    day_offset: int | None,
) -> str:
    if schedule_type is ScheduleType.HOURLY:
        return f"{minute_offset} * * * *"
    elif schedule_type is ScheduleType.DAILY:
        return f"{minute_offset} {hour_offset} * * *"
    elif schedule_type is ScheduleType.WEEKLY:
        return f"{minute_offset} {hour_offset} * * {day_offset if day_offset is not None else 0}"
    elif schedule_type is ScheduleType.MONTHLY:
        return f"{minute_offset} {hour_offset} {day_offset if day_offset is not None else 1} * *"
    else:
        check.assert_never(schedule_type)



import dataclasses
import importlib
from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, TypeVar, Union, overload

from dagster_shared import check
from dagster_shared.yaml_utils.source_position import SourcePositionTree
from typing_extensions import Self

from dagster._annotations import PublicAttr, deprecated, public
from dagster._core.definitions.definitions_class import Definitions
from dagster._utils import pushd
from dagster.components.resolved.context import ResolutionContext

if TYPE_CHECKING:
    from dagster.components.component.component import Component
    from dagster.components.core.component_tree import ComponentTree
    from dagster.components.core.decl import ComponentDecl
    from dagster.components.core.defs_module import ComponentPath, ResolvableToComponentPath


RESOLUTION_CONTEXT_STASH_KEY = "component_load_context"

T = TypeVar("T")


@public
@dataclass(frozen=True)
class ComponentDeclLoadContext:
    """Context object that provides environment and path information during loading of
    ComponentDecls. This context is automatically created and passed to ComponentDecl
    objects when loading a project's defs folder. Each Python module or folder in the
    defs directory receives a unique context instance that provides access to project
    structure, paths, and utilities.

    Args:
        path: The filesystem path of the component decl currently being loaded.
            For a file: ``/path/to/project/src/project/defs/my_component.py``
            For a directory: ``/path/to/project/src/project/defs/my_component/``
        project_root: The root directory of the Dagster project, typically containing
            ``pyproject.toml`` or ``setup.py``. Example: ``/path/to/project``
        defs_module_path: The filesystem path to the root defs folder.
            Example: ``/path/to/project/src/project/defs``
        defs_module_name: The Python module name for the root defs folder, used for
            import resolution. Typically follows the pattern ``"project_name.defs"``.
            Example: ``"my_project.defs"``
        resolution_context: The resolution context used by the component templating
            system for parameter resolution and variable substitution.
        component_tree: The component tree that contains the component decl currently being loaded.
        terminate_autoloading_on_keyword_files: Controls whether autoloading stops
            when encountering ``definitions.py`` or ``component.py`` files.
            **Deprecated**: This parameter will be removed after version 1.11.

                )

    Note:
        This context is automatically provided by Dagster's autoloading system and
        should not be instantiated manually in most cases. For testing purposes,
        use ``ComponentTree.for_test().decl_load_context`` to create a test instance.

    See Also:
        - :py:func:`dagster.definitions`: Decorator that receives this context
        - :py:class:`dagster.Definitions`: The object typically returned by context-using functions
        - :py:class:`dagster.components.resolved.context.ResolutionContext`: Underlying resolution context
        - :py:class:`dagster.ComponentLoadContext`: Context available when instantiating Components
    """

    project_root: PublicAttr[Path]
    defs_module_path: PublicAttr[Path]
    defs_module_name: PublicAttr[str]
    resolution_context: PublicAttr[ResolutionContext]
    component_tree: "ComponentTree"
    terminate_autoloading_on_keyword_files: bool
    component_path: PublicAttr["ComponentPath"]

    @property
    def path(self) -> Path:
        return self.component_path.file_path

    def __post_init__(self):
        object.__setattr__(
            self,
            "resolution_context",
            self.resolution_context.with_stashed_value(RESOLUTION_CONTEXT_STASH_KEY, self),
        )

    @staticmethod
    def from_resolution_context(
        resolution_context: ResolutionContext,
    ) -> "ComponentDeclLoadContext":
        return check.inst(
            resolution_context.stash.get(RESOLUTION_CONTEXT_STASH_KEY), ComponentDeclLoadContext
        )

    def _with_resolution_context(self, resolution_context: ResolutionContext) -> "Self":
        return dataclasses.replace(self, resolution_context=resolution_context)

    def with_rendering_scope(self, rendering_scope: Mapping[str, Any]) -> "Self":
        return self._with_resolution_context(self.resolution_context.with_scope(**rendering_scope))

    def with_source_position_tree(self, source_position_tree: SourcePositionTree) -> "Self":
        return self._with_resolution_context(
            self.resolution_context.with_source_position_tree(source_position_tree)
        )

    def for_component_path(self, component_path: "ComponentPath") -> "Self":
        return dataclasses.replace(self, component_path=component_path)

    def defs_relative_module_name(self, path: Path) -> str:
        """Returns the name of the python module at the given path, relative to the project root."""
        container_path = self.path.parent if self.path.is_file() else self.path
        with pushd(str(container_path)):
            relative_path = path.resolve().relative_to(self.defs_module_path.resolve())
            if path.name == "__init__.py":
                # e.g. "a_project/defs/something/__init__.py" -> "a_project.defs.something"
                relative_parts = relative_path.parts[:-1]
            elif path.is_file():
                # e.g. "a_project/defs/something/file.py" -> "a_project.defs.something.file"
                relative_parts = [*relative_path.parts[:-1], relative_path.stem]
            else:
                # e.g. "a_project/defs/something/" -> "a_project.defs.something"
                relative_parts = relative_path.parts
            return ".".join([self.defs_module_name, *relative_parts])

    def normalize_component_type_str(self, type_str: str) -> str:
        return (
            f"{self.defs_relative_module_name(self.path)}{type_str}"
            if type_str.startswith(".")
            else type_str
        )

    def load_defs_relative_python_module(self, path: Path) -> ModuleType:
        """Load a python module relative to the defs's context path. This is useful for loading code
        the resides within the defs directory.

        Example:
            .. code-block:: python

                def build_defs(self, context: ComponentLoadContext) -> Definitions:
                    return load_definitions_from_module(
                        context.load_defs_relative_python_module(
                            Path(self.definitions_path) if self.definitions_path else Path("definitions.py")
                        )
                    )

        In a typical setup you end up with module names such as "a_project.defs.my_component.my_python_file".

        This handles "__init__.py" files by ending the module name at the parent directory
        (e.g "a_project.defs.my_component") if the file resides at "a_project/defs/my_component/__init__.py".

        This calls importlib.import_module with that module name, going through the python module import system.

        It is as if one typed "import a_project.defs.my_component.my_python_file" in the python interpreter.
        """
        return importlib.import_module(self.defs_relative_module_name(path))

    @overload
    def load_component(self, defs_path: "ResolvableToComponentPath") -> "Component": ...
    @overload
    def load_component(
        self, defs_path: "ResolvableToComponentPath", expected_type: type[T]
    ) -> T: ...

    def load_component(
        self, defs_path: "ResolvableToComponentPath", expected_type: type[T] | None = None
    ) -> Any:
        """Loads a component from the given path.

        Args:
            defs_path: Path to the component to load. If relative, resolves relative to the defs root.

        Returns:
            Component: The component loaded from the given path.
        """
        from dagster.components.core.defs_module import ComponentPath

        resolved_path = ComponentPath.from_resolvable(self.defs_module_path, defs_path)
        self.component_tree.mark_component_load_dependency(
            from_path=self.component_path, to_path=resolved_path
        )
        return self.component_tree.load_component(resolved_path, expected_type)  # type: ignore[reportIncompatibleArgumentType]

    @overload
    def load_component_at_path(self, defs_path: "ResolvableToComponentPath") -> "Component": ...
    @overload
    def load_component_at_path(
        self, defs_path: "ResolvableToComponentPath", expected_type: type[T]
    ) -> T: ...

    @deprecated(
        breaking_version="1.13.0",
        additional_warn_text="Use load_component instead.",
    )
    def load_component_at_path(
        self, defs_path: "ResolvableToComponentPath", expected_type: type[T] | None = None
    ) -> Any:
        """Loads a component from the given path.

        Args:
            defs_path: Path to the component to load. If relative, resolves relative to the defs root.

        Returns:
            Component: The component loaded from the given path.

        """
        return self.load_component(defs_path, expected_type)  # type: ignore[reportIncompatibleArgumentType]

    def load_structural_component_at_path(
        self, defs_path: "ResolvableToComponentPath"
    ) -> "Component":
        """Loads a component from the given path.

        Args:
            defs_path: Path to the component to load. If relative, resolves relative to the defs root.

        Returns:
            Component: The component loaded from the given path.
        """
        from dagster.components.core.defs_module import ComponentPath

        resolved_path = ComponentPath.from_resolvable(self.defs_module_path, defs_path)
        self.component_tree.mark_component_load_dependency(
            from_path=self.component_path, to_path=resolved_path
        )
        return self.component_tree.load_structural_component_at_path(resolved_path)


@public
@dataclass(frozen=True)
class ComponentLoadContext(ComponentDeclLoadContext):
    """Context object that provides environment and path information during component loading.
    This context is automatically created and passed to component definitions when loading
    a project's defs folder. Each Python module or folder in the defs directory receives
    a unique context instance that provides access to the underlying ComponentDecl,
    project structure, paths, and utilities for dynamic component instantiation.

    The context enables components to:
    - Access project and module path information
    - Load other modules and definitions within the project
    - Resolve relative imports and module names
    - Access templating and resolution capabilities

    Args:
        path: The filesystem path of the component currently being loaded.
            For a file: ``/path/to/project/src/project/defs/my_component.py``
            For a directory: ``/path/to/project/src/project/defs/my_component/``
        project_root: The root directory of the Dagster project, typically containing
            ``pyproject.toml`` or ``setup.py``. Example: ``/path/to/project``
        defs_module_path: The filesystem path to the root defs folder.
            Example: ``/path/to/project/src/project/defs``
        defs_module_name: The Python module name for the root defs folder, used for
            import resolution. Typically follows the pattern ``"project_name.defs"``.
            Example: ``"my_project.defs"``
        resolution_context: The resolution context used by the component templating
            system for parameter resolution and variable substitution.
        component_tree: The component tree that contains the component currently being loaded.
        terminate_autoloading_on_keyword_files: Controls whether autoloading stops
            when encountering ``definitions.py`` or ``component.py`` files.
            **Deprecated**: This parameter will be removed after version 1.11.
        component_decl: The associated ComponentDecl to the component being loaded.

    Note:
        This context is automatically provided by Dagster's autoloading system and
        should not be instantiated manually in most cases. For testing purposes,
        use ``ComponentTree.for_test().load_context`` to create a test instance.

    See Also:
        - :py:func:`dagster.definitions`: Decorator that receives this context
        - :py:class:`dagster.Definitions`: The object typically returned by context-using functions
        - :py:class:`dagster.components.resolved.context.ResolutionContext`: Underlying resolution context
        - :py:class:`dagster.ComponentDeclLoadContext`: Context available when loading ComponentDecls
    """

    component_decl: "ComponentDecl"

    @staticmethod
    def from_decl_load_context(
        decl_load_context: ComponentDeclLoadContext,
        component_decl: "ComponentDecl",
    ) -> "ComponentLoadContext":
        """Augments a ComponentDeclLoadContext with the ComponentDecl being loaded.

        Args:
            decl_load_context: The ComponentDeclLoadContext to augment.
            component_decl: The ComponentDecl being loaded.

        Returns:
            ComponentLoadContext: An augmented context which can be used to
            load and build definitions for the component.
        """
        return ComponentLoadContext(
            component_path=decl_load_context.component_path,
            project_root=decl_load_context.project_root,
            defs_module_path=decl_load_context.defs_module_path,
            defs_module_name=decl_load_context.defs_module_name,
            resolution_context=decl_load_context.resolution_context,
            component_tree=decl_load_context.component_tree,
            terminate_autoloading_on_keyword_files=decl_load_context.terminate_autoloading_on_keyword_files,
            component_decl=component_decl,
        )

    def build_defs(self, defs_path: Union[Path, "ComponentPath"]) -> Definitions:
        """Builds definitions from the given defs subdirectory. Currently
        does not incorporate postprocessing from parent defs modules.

        Args:
            defs_path: Path to the defs module to load. If relative, resolves relative to the defs root.

        Returns:
            Definitions: The definitions loaded from the given path.
        """
        from dagster.components.core.defs_module import ComponentPath

        resolved_path = ComponentPath.from_resolvable(self.defs_module_path, defs_path)
        self.component_tree.mark_component_defs_dependency(
            from_path=self.component_path, to_path=resolved_path
        )
        return self.component_tree.build_defs(resolved_path)

    @deprecated(
        breaking_version="1.13.0",
        additional_warn_text="Use build_defs instead.",
    )
    def build_defs_at_path(self, defs_path: Union[Path, "ComponentPath"]) -> Definitions:
        """Builds definitions from the given defs subdirectory. Currently
        does not incorporate postprocessing from parent defs modules.

        Args:
            defs_path: Path to the defs module to load. If relative, resolves relative to the defs root.

        Returns:
            Definitions: The definitions loaded from the given path.

        """
        return self.build_defs(defs_path)

    def for_path(self, path: Path) -> "Self":
        """Creates a new context for the given path.

        Args:
            path: The filesystem path to create a new context for.

        Returns:
            ComponentLoadContext: A new context for the given path.
        """
        from dagster.components.core.defs_module import ComponentPath

        component_path = ComponentPath.from_path(path=path)
        return self.for_component_path(component_path)



import inspect
from collections.abc import Callable, Generator
from typing import TYPE_CHECKING, Any, TypeAlias, overload

from dagster._symbol_annotations.public import public

if TYPE_CHECKING:
    from dagster.components.core.context import ComponentDeclLoadContext

# Template variable functions must have 0 or 1 parameters.
# When 1 parameter is present, it must be ComponentLoadContext.
TemplateVarFn: TypeAlias = Callable[..., Any]
"""Type alias for template variable functions.

Template variable functions are constrained to two specific signatures:

**Valid Signatures:**
    - ``() -> Any`` - Zero parameters for static values computed at load time
    - ``(context: ComponentLoadContext) -> Any`` - Single context parameter for context-aware values

**Type Constraints:**
    - Must have exactly 0 or 1 parameters (no more, no less)
    - If 1 parameter is present, it must be typed as ``ComponentLoadContext``
    - Return type can be any Python object (primitives, collections, objects, callables)
    - When used as ``@staticmethod``, same parameter constraints apply

**Usage Examples:**
    .. code-block:: python
    
        # ✅ Valid: Zero parameters (static value)
        def database_url() -> str: ...
        
        # ✅ Valid: Single ComponentLoadContext parameter
        def context_table_name(context: ComponentLoadContext) -> str: ...
        
        # ✅ Valid: Returns callable for Jinja2 UDF usage
        def table_generator() -> Callable[[str], str]: ...
        
        # ❌ Invalid: Multiple parameters
        def invalid_multi(context: ComponentLoadContext, other: str) -> str: ...
        
        # ❌ Invalid: Wrong context type
        def invalid_context(context: ComponentDeclLoadContext) -> str: ...

Note:
    While the type alias uses ``Callable[..., Any]`` for compatibility with Python's
    type system (particularly ``staticmethod``), the actual runtime validation enforces
    the strict 0-or-1 parameter constraint documented above.
"""

TEMPLATE_VAR_ATTR = "__dagster_template_var"


@overload
def template_var(fn: TemplateVarFn) -> TemplateVarFn: ...


@overload
def template_var() -> Callable[[TemplateVarFn], TemplateVarFn]: ...


@public
def template_var(
    fn: TemplateVarFn | None = None,
) -> TemplateVarFn | Callable[[TemplateVarFn], TemplateVarFn]:
    """Decorator that marks a function as a template variable for use in component YAML definitions.

    Template variables provide dynamic values and functions that can be injected into component
    YAML definitions using Jinja2 templating syntax ({{ variable_name }}). They are evaluated
    at component load time and can optionally receive a ComponentLoadContext parameter for
    context-aware behavior.

    These values can be any python object and are passed directly to the component as Python object.
    They can be injected at any level of the defs file.

    There are two main usage patterns:

    1. **Module-level template variables**: Functions defined in a separate module and referenced
       via the ``template_vars_module`` field in component YAML
    2. **Component class static methods**: Template variables defined as ``@staticmethod`` on
       a Component class, automatically available to instances of that component

    Template vars can themselves be functions, in which case they are user-defined functions, invoked
    with function syntax within the defs file.

    Args:
        fn: The function to decorate as a template variable. If None, returns a decorator.

    Returns:
        The decorated function with template variable metadata, or a decorator function.

    Note:
        Template variables are evaluated at component load time, not at runtime. They provide
        configuration values and functions for YAML templating, not runtime component logic.

    Function Signatures:
        Template variable functions can have one of two valid signatures:

        **Zero parameters (static values)**:

        .. code-block:: python

            @dg.template_var
            def static_value() -> Any:
                # Returns a static value computed at load time
                return "computed_value"

        **Single ComponentLoadContext parameter (context-aware)**:

        .. code-block:: python

            @dg.template_var
            def context_value(context: dg.ComponentLoadContext) -> Any:
                # Returns a value based on the component's loading context
                return f"value_{context.path.name}"

        **Return Types:**
        Template variables can return any type, including:

        - **Primitive values**: ``str``, ``int``, ``bool``, ``float``
        - **Collections**: ``list``, ``dict``, ``set``, ``tuple``
        - **Complex objects**: ``PartitionsDefinition``, custom classes, etc.
        - **Functions**: ``Callable`` objects for use as UDFs in Jinja2 templates

        **Invalid Signatures:**

        .. code-block:: python

            # ❌ Multiple parameters not allowed
            @dg.template_var
            def invalid_multiple_params(context: ComponentLoadContext, other_param: str):
                pass

            # ❌ Wrong context type
            @dg.template_var
            def invalid_context_type(context: ComponentDeclLoadContext):
                pass

            # ❌ Static methods with parameters other than context
            class MyComponent(dg.Component):
                @staticmethod
                @dg.template_var
                def invalid_static(param: str):  # Only 0 or 1 (context) params allowed
                    pass

    Examples:
        **Basic template variable (no context needed)**:

        .. code-block:: python

            import dagster as dg
            import os

            @dg.template_var
            def database_url() -> str:
                if os.getenv("ENVIRONMENT") == "prod":
                    return "postgresql://prod-server:5432/db"
                else:
                    return "postgresql://localhost:5432/dev_db"

        **Context-aware template variable**:

        .. code-block:: python

            @dg.template_var
            def component_specific_table(context: dg.ComponentLoadContext) -> str:
                return f"table_{context.path.name}"

        **Template variable returning a function**:

        This is colloquially called a "udf" (user-defined function).

        .. code-block:: python

            @dg.template_var
            def table_name_generator() -> Callable[[str], str]:
                return lambda prefix: f"{prefix}_processed_data"

        **Using template variables in YAML**:

        .. code-block:: yaml

            # defs.yaml
            type: my_project.components.DataProcessor
            template_vars_module: .template_vars
            attributes:
              database_url: "{{ database_url }}"
              table_name: "{{ component_specific_table }}"
              processed_table: "{{ table_name_generator('sales') }}"

        **Component class static methods**:

        .. code-block:: python

            class MyComponent(dg.Component):
                @staticmethod
                @dg.template_var
                def default_config() -> dict:
                    return {"timeout": 30, "retries": 3}

                @staticmethod
                @dg.template_var
                def context_aware_value(context: dg.ComponentLoadContext) -> str:
                    return f"value_for_{context.path.name}"

        **Using in YAML (component static methods)**:

        .. code-block:: yaml

            type: my_project.components.MyComponent
            attributes:
              config: "{{ default_config }}"
              name: "{{ context_aware_value }}"

    See Also:
        - :py:class:`dagster.ComponentLoadContext`: Context object available to template variables
    """

    def decorator(func: TemplateVarFn) -> TemplateVarFn:
        setattr(func, TEMPLATE_VAR_ATTR, True)
        return func

    if fn is None:
        return decorator
    else:
        return decorator(fn)


def is_template_var(obj: Any) -> bool:
    return getattr(obj, TEMPLATE_VAR_ATTR, False)


def find_inline_template_vars_in_module(module: Any) -> dict[str, TemplateVarFn]:
    """Finds all template functions in the given module.

    Args:
        module: The module to search for template functions

    Returns:
        dict[str, TemplateVarFn]: A dictionary of template variable functions indexed by name
    """
    return {name: obj for name, obj in inspect.getmembers(module, is_template_var)}


def is_staticmethod(cls, method):
    """Check if a method is a static method by examining the class descriptor.

    Args:
        cls: The class where the method is defined
        method: The method reference (e.g., cls.method_name)

    Returns:
        bool: True if the method is a static method, False otherwise
    """
    # Get the method name from the method reference
    if not hasattr(method, "__name__"):
        return False

    method_name = method.__name__

    # Check the class hierarchy to find where this method is defined
    for base_cls in inspect.getmro(cls):
        if method_name in base_cls.__dict__:
            method_descriptor = base_cls.__dict__[method_name]
            return isinstance(method_descriptor, staticmethod)

    return False


def _get_all_static_template_vars(
    cls: type,
) -> Generator[tuple[str, TemplateVarFn, inspect.Signature], None, None]:
    """Helper function to find all staticmethods that are template variables.

    Args:
        cls: The class to search through

    Yields:
        Tuple of (name, func, signature) for each static template variable found
    """
    for name, method in cls.__dict__.items():
        if is_staticmethod(cls, method):
            # Get the actual function from the staticmethod wrapper
            func = method.__get__(None, cls)
            if is_template_var(func):
                sig = inspect.signature(func)
                if len(sig.parameters) > 1:
                    raise ValueError(
                        f"Static template var {name} must have 0 or 1 parameters, got {len(sig.parameters)}"
                    )
                yield name, func, sig


def get_context_free_static_template_vars(cls: type) -> dict[str, TemplateVarFn]:
    """Find all staticmethods in a class that can be used as template variables that do not have a context parameter.
    This is a separate function because of legacy reasons. It is the default implementation of get_additional_scope.

    Args:
        cls: The class to search through

    Returns:
        A dictionary mapping method names to their callable functions for staticmethods that don't require context
    """
    results = {}
    for name, func, sig in _get_all_static_template_vars(cls):
        if len(sig.parameters) == 0:
            results[name] = func()

    return results


def get_context_aware_static_template_vars(
    cls: type, context: "ComponentDeclLoadContext"
) -> dict[str, TemplateVarFn]:
    """Find staticmethods that require context and invoke them with the provided context.

    Args:
        cls: The class to search through
        context: ComponentDeclLoadContext to pass to template variables

    Returns:
        A dictionary mapping method names to their values for context-requiring staticmethods
    """
    results = {}
    for name, func, sig in _get_all_static_template_vars(cls):
        if len(sig.parameters) == 1:
            results[name] = func(context)

    return results



from typing import Any, Generic

from typing_extensions import TypeVar

import dagster._check as check
from dagster._annotations import deprecated

T_cov = TypeVar("T_cov", default=Any, covariant=True)


@deprecated(breaking_version="2.0", additional_warn_text="Use string partition keys instead.")
class Partition(Generic[T_cov]):
    """A Partition represents a single slice of the entire set of a job's possible work. It consists
    of a value, which is an object that represents that partition, and an optional name, which is
    used to label the partition in a human-readable way.

    Args:
        value (Any): The object for this partition
        name (str): Name for this partition
    """

    def __init__(self, value: Any, name: str | None = None):
        self._value = value
        self._name = check.str_param(name or str(value), "name")

    @property
    def value(self) -> T_cov:
        return self._value

    @property
    def name(self) -> str:
        return self._name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Partition):
            return False
        else:
            return self.value == other.value and self.name == other.name

    __hash__ = None  # pyright: ignore[reportAssignmentType]



import os
from collections.abc import Mapping
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING

import dagster._check as check
from dagster._core.errors import DagsterUserCodeProcessError
from dagster._core.remote_representation.external_data import RepositoryErrorSnap, RepositorySnap
from dagster._serdes import deserialize_value
from dagster._utils.error import SerializableErrorInfo

if TYPE_CHECKING:
    from dagster._core.remote_representation.code_location import CodeLocation
    from dagster._core.snap import JobSnap
    from dagster._grpc.client import DagsterGrpcClient


def sync_get_external_repositories_data_grpc(
    api_client: "DagsterGrpcClient", code_location: "CodeLocation", defer_snapshots: bool
) -> Mapping[str, tuple[RepositorySnap, Mapping[str, "JobSnap"]]]:
    from dagster._core.remote_origin import RemoteRepositoryOrigin
    from dagster._core.remote_representation.code_location import CodeLocation
    from dagster._core.remote_representation.external_data import (
        extract_serialized_job_snap_from_serialized_job_data_snap,
    )
    from dagster._core.snap import JobSnap

    check.inst_param(code_location, "code_location", CodeLocation)

    repo_datas = {}
    for repository_name in code_location.repository_names:  # type: ignore
        repo_origin = RemoteRepositoryOrigin(
            code_location.origin,
            repository_name,
        )
        result = deserialize_value(
            api_client.external_repository(
                remote_repository_origin=RemoteRepositoryOrigin(
                    code_location.origin,
                    repository_name,
                ),
                defer_snapshots=defer_snapshots,
            ),
            (RepositorySnap, RepositoryErrorSnap),
        )

        if isinstance(result, RepositoryErrorSnap):
            raise DagsterUserCodeProcessError.from_error_info(result.error)

        job_snaps = {}

        if not result.has_job_data():
            job_refs = result.get_job_refs()
            with ThreadPoolExecutor(
                max_workers=int(os.getenv("DAGSTER_EXTERNAL_JOB_THREADPOOL_WORKERS", "4"))
            ) as executor:
                job_responses = list(
                    executor.map(
                        lambda job_ref: api_client.external_job(
                            repo_origin,
                            job_ref.name,
                        ),
                        job_refs,
                    )
                )

            for i, job_ref in enumerate(job_refs):
                job_response = job_responses[i]
                if not job_response.serialized_job_data:
                    error = (
                        deserialize_value(job_response.serialized_error, SerializableErrorInfo)
                        if job_response.serialized_error
                        else "no captured error"
                    )
                    raise Exception(
                        f"Error fetching job data for {job_ref.name} in code server:\n{error}"
                    )
                job_snap = deserialize_value(
                    extract_serialized_job_snap_from_serialized_job_data_snap(
                        job_response.serialized_job_data
                    ),
                    JobSnap,
                )

                job_snaps[job_ref.name] = job_snap

        repo_datas[repository_name] = result, job_snaps
    return repo_datas



from collections.abc import Sequence
from datetime import datetime
from typing import TYPE_CHECKING, Optional

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.partitions.context import (
    PartitionLoadingContext,
    partition_loading_context,
)
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.utils.base import (
    raise_error_on_duplicate_partition_keys,
    raise_error_on_invalid_partition_key_substring,
)
from dagster._core.types.pagination import PaginatedResults

if TYPE_CHECKING:
    from dagster._core.instance import DynamicPartitionsStore


class StaticPartitionsDefinition(PartitionsDefinition[str]):
    """A statically-defined set of partitions.

    We recommended limiting partition counts for each asset to 100,000 partitions or fewer.

    Example:
        .. code-block:: python

            from dagster import StaticPartitionsDefinition, asset

            oceans_partitions_def = StaticPartitionsDefinition(
                ["arctic", "atlantic", "indian", "pacific", "southern"]
            )

            @asset(partitions_def=oceans_partitions_defs)
            def ml_model_for_each_ocean():
                ...
    """

    def __init__(self, partition_keys: Sequence[str]):
        # for back compat reasons we allow str as a Sequence[str] here
        if not isinstance(partition_keys, str):
            check.sequence_param(
                partition_keys,
                "partition_keys",
                of_type=str,
            )

        raise_error_on_invalid_partition_key_substring(partition_keys)
        raise_error_on_duplicate_partition_keys(partition_keys)

        self._partition_keys = partition_keys

    @public
    def get_partition_keys(
        self,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> Sequence[str]:
        """Returns a list of strings representing the partition keys of the PartitionsDefinition.

        Args:
            current_time (Optional[datetime]): A datetime object representing the current time, only
                applicable to time-based partitions definitions.
            dynamic_partitions_store (Optional[DynamicPartitionsStore]): The DynamicPartitionsStore
                object that is responsible for fetching dynamic partitions. Only applicable to
                DynamicPartitionsDefinitions.

        Returns:
            Sequence[str]

        """
        return self._partition_keys

    def get_paginated_partition_keys(
        self,
        context: PartitionLoadingContext,
        limit: int,
        ascending: bool,
        cursor: str | None = None,
    ) -> PaginatedResults[str]:
        with partition_loading_context(new_ctx=context):
            partition_keys = self.get_partition_keys()
            return PaginatedResults.create_from_sequence(
                partition_keys, limit=limit, ascending=ascending, cursor=cursor
            )

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other) -> bool:
        return isinstance(other, StaticPartitionsDefinition) and (
            self is other or self._partition_keys == other.get_partition_keys()
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}(partition_keys={self._partition_keys})"

    def get_num_partitions(self) -> int:
        # We don't currently throw an error when a duplicate partition key is defined
        # in a static partitions definition, though we will at 1.3.0.
        # This ensures that partition counts are correct in the Dagster UI.
        return len(set(self.get_partition_keys()))



import operator
from collections.abc import Callable, Sequence
from dataclasses import dataclass, replace
from typing import Any, Generic, Optional, TypeAlias

from dagster_shared.serdes.serdes import DataclassSerializer, whitelist_for_serdes
from typing_extensions import Self

import dagster._check as check
from dagster._core.definitions.asset_key import T_EntityKey
from dagster._core.definitions.events import AssetKeyPartitionKey
from dagster._core.definitions.partitions.context import partition_loading_context
from dagster._core.definitions.partitions.definition import PartitionsDefinition
from dagster._core.definitions.partitions.snap.snap import PartitionsSnap
from dagster._core.definitions.partitions.subset import (
    AllPartitionsSubset,
    DefaultPartitionsSubset,
    PartitionsSubset,
    TimeWindowPartitionsSubset,
)
from dagster._core.definitions.partitions.subset.key_ranges import KeyRangesPartitionsSubset

EntitySubsetValue: TypeAlias = bool | PartitionsSubset

CoercibleToAssetEntitySubsetValue: TypeAlias = str | Sequence[str] | PartitionsSubset | None


class EntitySubsetSerializer(DataclassSerializer):
    """Ensures that the inner PartitionsSubset is converted to a serializable form if necessary."""

    def get_storage_name(self) -> str:
        # backcompat
        return "AssetSubset"

    def before_pack(self, value: "SerializableEntitySubset") -> "SerializableEntitySubset":  # pyright: ignore[reportIncompatibleMethodOverride]
        if value.is_partitioned:
            return replace(value, value=value.subset_value.to_serializable_subset())
        return value


@whitelist_for_serdes(
    serializer=EntitySubsetSerializer,
    storage_field_names={"key": "asset_key"},
    old_storage_names={"AssetSubset"},
)
@dataclass(frozen=True)
class SerializableEntitySubset(Generic[T_EntityKey]):
    """Represents a serializable subset of a given EntityKey."""

    key: T_EntityKey
    value: EntitySubsetValue

    @classmethod
    def empty(
        cls, key: T_EntityKey, partitions_def: PartitionsDefinition | None
    ) -> "SerializableEntitySubset[T_EntityKey]":
        return cls(key=key, value=partitions_def.empty_subset() if partitions_def else False)

    @classmethod
    def from_coercible_value(
        cls,
        key: T_EntityKey,
        value: CoercibleToAssetEntitySubsetValue,
        partitions_def: PartitionsDefinition | None,
    ) -> "SerializableEntitySubset":
        """Creates a new SerializableEntitySubset, handling coercion of a CoercibleToAssetEntitySubsetValue
        to an EntitySubsetValue.
        """
        if value is None:
            check.invariant(
                partitions_def is None,
                "Cannot create a SerializableEntitySubset with value=None and non-None partitions_def",
            )
            return cls(key=key, value=True)
        if isinstance(value, str):
            partitions_def = check.not_none(partitions_def)
            if partitions_def.partitions_subset_class is not DefaultPartitionsSubset:
                # DefaultPartitionsSubset just adds partition keys to a set, but other subsets
                # may require partition keys be part of the partition, so validate the key
                with partition_loading_context() as ctx:
                    partitions_def.validate_partition_key(value, context=ctx)
            partitions_subset = partitions_def.subset_with_partition_keys([value])
        elif isinstance(value, PartitionsSubset):
            if partitions_def is not None:
                check.inst_param(
                    value,
                    "value",
                    partitions_def.partitions_subset_class,
                )
            partitions_subset = value
        else:
            check.list_param(value, "value", of_type=str)
            partitions_def = check.not_none(partitions_def)
            partitions_subset = partitions_def.subset_with_partition_keys(value)
        return cls(key=key, value=partitions_subset)

    @classmethod
    def try_from_coercible_value(
        cls,
        key: T_EntityKey,
        value: CoercibleToAssetEntitySubsetValue,
        partitions_def: PartitionsDefinition | None,
    ) -> Optional["SerializableEntitySubset"]:
        """Attempts to create a new SerializableEntitySubset, handling coercion of a CoercibleToAssetEntitySubsetValue
        and partitions definition to an EntitySubsetValue. Returns None if the coercion fails.
        """
        try:
            return cls.from_coercible_value(key, value, partitions_def)
        except:
            return None

    @property
    def is_partitioned(self) -> bool:
        return not isinstance(self.value, bool)

    @property
    def bool_value(self) -> bool:
        return check.inst(self.value, bool)

    @property
    def subset_value(self) -> PartitionsSubset:
        return check.inst(self.value, PartitionsSubset)

    @property
    def size(self) -> int:
        if not self.is_partitioned:
            return int(self.bool_value)
        else:
            return len(self.subset_value)

    @property
    def is_empty(self) -> bool:
        if self.is_partitioned:
            return self.subset_value.is_empty
        else:
            return not self.bool_value

    def is_compatible_with_partitions_def(
        self, partitions_def: PartitionsDefinition | None
    ) -> bool:
        from dagster._core.definitions.partitions.definition.time_window import (
            TimeWindowPartitionsDefinition,
        )

        if self.is_partitioned:
            # for some PartitionSubset types, we have access to the underlying partitions
            # definitions, so we can ensure those are identical
            if isinstance(self.value, (TimeWindowPartitionsSubset, AllPartitionsSubset)):
                return self.value.partitions_def == partitions_def
            # for KeyRangesPartitionsSubset, we have the PartitionsSnap, so we can use that
            elif isinstance(self.value, KeyRangesPartitionsSubset):
                if (
                    partitions_def is None
                    or PartitionsSnap.from_def(partitions_def) != self.value.partitions_snap
                ):
                    return False
                # all ranges must be valid
                return all(
                    partitions_def.has_partition_key(r.start)
                    and partitions_def.has_partition_key(r.end)
                    for r in self.value.key_ranges
                )
            elif isinstance(self.value, DefaultPartitionsSubset) and isinstance(
                partitions_def, TimeWindowPartitionsDefinition
            ):
                return all(partitions_def.has_partition_key(k) for k in self.value.subset)

            else:
                return partitions_def is not None
        else:
            return partitions_def is None

    def _oper(self, other: Self, oper: Callable[..., Any]) -> Self:
        check.invariant(self.key == other.key, "Keys must match for operation")
        value = oper(self.value, other.value)
        return self.__class__(key=self.key, value=value)

    def compute_difference(self, other: Self) -> Self:
        if isinstance(self.value, bool):
            value = self.bool_value and not other.bool_value
            return self.__class__(key=self.key, value=value)
        else:
            return self._oper(other, operator.sub)

    def compute_union(self, other: Self) -> Self:
        return self._oper(other, operator.or_)

    def compute_intersection(self, other: Self) -> Self:
        return self._oper(other, operator.and_)

    def __contains__(self, item: AssetKeyPartitionKey) -> bool:
        if not self.is_partitioned:
            return item.asset_key == self.key and item.partition_key is None and self.bool_value
        else:
            return item.asset_key == self.key and item.partition_key in self.subset_value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self.key}>({self.value})"



import os
import shlex
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Union

from dagster.components.lib.executable_component.component import OpSpec

if TYPE_CHECKING:
    from dagster._core.execution.context.asset_check_execution_context import (
        AssetCheckExecutionContext,
    )
    from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
    from dagster._core.pipes.context import PipesExecutionResult


class ScriptSpec(OpSpec):
    type: Literal["script"] = "script"
    path: str
    args: list[str] | str | None = None

    @staticmethod
    def with_script_stem_as_default_name(
        script_runner_spec: "ScriptSpec",
    ) -> "ScriptSpec":
        return script_runner_spec.model_copy(
            update={
                "name": script_runner_spec.name
                if script_runner_spec.name
                else Path(script_runner_spec.path).stem
            }
        )


def invoke_runner(
    *, context: Union["AssetExecutionContext", "AssetCheckExecutionContext"], command: list[str]
) -> Sequence["PipesExecutionResult"]:
    from dagster._core.pipes.subprocess import PipesSubprocessClient

    return (
        PipesSubprocessClient()
        .run(context=context.op_execution_context, command=command)
        .get_results()
    )


def get_cmd(script_runner_exe: list[str], spec: ScriptSpec, path: str) -> list[str]:
    abs_path = spec.path if os.path.isabs(spec.path) else os.path.join(path, spec.path)
    if isinstance(spec.args, str):
        return [*script_runner_exe, abs_path, *shlex.split(spec.args)]
    elif isinstance(spec.args, list):
        return [*script_runner_exe, abs_path, *spec.args]
    else:
        return [*script_runner_exe, abs_path]



"""Event methods implementation - consolidated from EventDomain."""

import logging
import logging.config
import sys
import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional, Union

import dagster._check as check
from dagster._annotations import deprecated
from dagster._time import get_current_timestamp
from dagster._utils.warnings import beta_warning

if TYPE_CHECKING:
    from dagster._core.event_api import EventHandlerFn
    from dagster._core.events import (
        DagsterEvent,
        DagsterEventBatchMetadata,
        DagsterEventType,
        EngineEventData,
        JobFailureData,
    )
    from dagster._core.events.log import EventLogEntry
    from dagster._core.instance.instance import DagsterInstance
    from dagster._core.instance.types import _EventListenerLogHandler
    from dagster._core.storage.dagster_run import DagsterRun
    from dagster._core.storage.event_log.base import (
        EventLogConnection,
        EventLogRecord,
        EventRecordsFilter,
    )


class EventMethods:
    """Mixin class containing event-related functionality for DagsterInstance.

    This class provides methods for event storage, querying, and streaming.
    All methods are implemented as instance methods that DagsterInstance inherits.
    """

    @property
    def _instance(self) -> "DagsterInstance":
        """Cast self to DagsterInstance for type-safe access to instance methods and properties."""
        from dagster._core.instance.instance import DagsterInstance

        return check.inst(self, DagsterInstance)

    # Private member access wrappers with consolidated type: ignore
    @property
    def _event_storage_impl(self):
        """Access to event storage."""
        return self._instance._event_storage  # noqa: SLF001

    @property
    def _event_buffer_impl(self):
        """Access to event buffer."""
        return self._instance._event_buffer  # noqa: SLF001

    @property
    def _run_storage_impl(self):
        """Access to run storage."""
        return self._instance._run_storage  # noqa: SLF001

    @property
    def _subscribers_impl(self):
        """Access to subscribers."""
        return self._instance._subscribers  # noqa: SLF001

    @property
    def _settings_impl(self):
        """Access to settings."""
        return self._instance._settings  # noqa: SLF001

    def logs_after(
        self,
        run_id: str,
        cursor: int | None = None,
        of_type: Optional["DagsterEventType"] = None,
        limit: int | None = None,
    ) -> Sequence["EventLogEntry"]:
        """Get logs after cursor."""
        return self._event_storage_impl.get_logs_for_run(
            run_id,
            cursor=cursor,
            of_type=of_type,
            limit=limit,
        )

    def all_logs(
        self,
        run_id: str,
        of_type: Union["DagsterEventType", set["DagsterEventType"]] | None = None,
    ) -> Sequence["EventLogEntry"]:
        """Get all logs for run."""
        return self._event_storage_impl.get_logs_for_run(run_id, of_type=of_type)

    def get_records_for_run(
        self,
        run_id: str,
        cursor: str | None = None,
        of_type: Union["DagsterEventType", set["DagsterEventType"]] | None = None,
        limit: int | None = None,
        ascending: bool = True,
    ) -> "EventLogConnection":
        """Get event records for run."""
        return self._event_storage_impl.get_records_for_run(
            run_id, cursor, of_type, limit, ascending
        )

    def watch_event_logs(self, run_id: str, cursor: str | None, cb: "EventHandlerFn") -> None:
        """Watch event logs."""
        return self._event_storage_impl.watch(run_id, cursor, cb)

    def end_watch_event_logs(self, run_id: str, cb: "EventHandlerFn") -> None:
        """End watch event logs."""
        return self._event_storage_impl.end_watch(run_id, cb)

    @deprecated(breaking_version="2.0")
    def get_event_records(
        self,
        event_records_filter: "EventRecordsFilter",
        limit: int | None = None,
        ascending: bool = False,
    ) -> Sequence["EventLogRecord"]:
        """Return a list of event records stored in the event log storage.

        Args:
            event_records_filter (Optional[EventRecordsFilter]): the filter by which to filter event
                records.
            limit (Optional[int]): Number of results to get. Defaults to infinite.
            ascending (Optional[bool]): Sort the result in ascending order if True, descending
                otherwise. Defaults to descending.

        Returns:
            List[EventLogRecord]: List of event log records stored in the event log storage.
        """
        from dagster._core.events import PIPELINE_EVENTS, DagsterEventType

        if (
            event_records_filter.event_type == DagsterEventType.ASSET_MATERIALIZATION_PLANNED
            and event_records_filter.asset_partitions
        ):
            warnings.warn(
                "Asset materialization planned events with partitions subsets will not be "
                "returned when the event records filter contains the asset_partitions argument"
            )
        elif event_records_filter.event_type == DagsterEventType.ASSET_MATERIALIZATION:
            warnings.warn(
                "Use fetch_materializations instead of get_event_records to fetch materialization events."
            )
        elif event_records_filter.event_type == DagsterEventType.ASSET_OBSERVATION:
            warnings.warn(
                "Use fetch_observations instead of get_event_records to fetch observation events."
            )
        elif event_records_filter.event_type in PIPELINE_EVENTS:
            warnings.warn(
                "Use fetch_run_status_changes instead of get_event_records to fetch run status change events."
            )

        return self._event_storage_impl.get_event_records(event_records_filter, limit, ascending)

    def should_store_event(self, event: "EventLogEntry") -> bool:
        """Check if event should be stored."""
        if (
            event.dagster_event is not None
            and event.dagster_event.is_asset_failed_to_materialize
            and not self._event_storage_impl.can_store_asset_failure_events
        ):
            return False
        return True

    def store_event(self, event: "EventLogEntry") -> None:
        """Store event."""
        if not self.should_store_event(event):
            return
        self._event_storage_impl.store_event(event)

    def _is_batch_writing_enabled(self) -> bool:
        """Check if batch writing is enabled."""
        return self._get_event_batch_size() > 0

    def _get_event_batch_size(self) -> int:
        """Get event batch size."""
        import os

        return int(os.getenv("DAGSTER_EVENT_BATCH_SIZE", "0"))

    def handle_new_event(
        self,
        event: "EventLogEntry",
        *,
        batch_metadata: Optional["DagsterEventBatchMetadata"] = None,
    ) -> None:
        """Handle a new event by storing it and notifying subscribers.

        Events may optionally be sent with `batch_metadata`. If batch writing is enabled, then
        events sent with `batch_metadata` will not trigger an immediate write. Instead, they will be
        kept in a batch-specific buffer (identified by `batch_metadata.id`) until either the buffer
        reaches the event batch size or the end of the batch is reached (signaled by
        `batch_metadata.is_end`). When this point is reached, all events in the buffer will be sent
        to the storage layer in a single batch. If an error occurrs during batch writing, then we
        fall back to iterative individual event writes.

        Args:
            event (EventLogEntry): The event to handle.
            batch_metadata (Optional[DagsterEventBatchMetadata]): Metadata for batch writing.
        """
        from dagster._core.events import RunFailureReason
        from dagster._core.storage.tags import RUN_FAILURE_REASON_TAG, WILL_RETRY_TAG
        from dagster._time import datetime_from_timestamp

        if not self.should_store_event(event):
            return

        if batch_metadata is None or not self._is_batch_writing_enabled():
            events = [event]
        else:
            batch_id, is_batch_end = batch_metadata.id, batch_metadata.is_end
            self._event_buffer_impl[batch_id].append(event)
            if (
                is_batch_end
                or len(self._event_buffer_impl[batch_id]) == self._get_event_batch_size()
            ):
                events = self._event_buffer_impl[batch_id]
                del self._event_buffer_impl[batch_id]
            else:
                return

        if len(events) == 1:
            self._event_storage_impl.store_event(events[0])
        else:
            try:
                self._event_storage_impl.store_event_batch(events)

            # Fall back to storing events one by one if writing a batch fails. We catch a generic
            # Exception because that is the parent class of the actually received error,
            # dagster_cloud_cli.core.errors.GraphQLStorageError, which we cannot import here due to
            # it living in a cloud package.
            except Exception as e:
                sys.stderr.write(f"Exception while storing event batch: {e}\n")
                sys.stderr.write(
                    "Falling back to storing multiple single-event storage requests...\n"
                )
                for event in events:
                    self._event_storage_impl.store_event(event)

        for event in events:
            run_id = event.run_id
            if (
                not self._event_storage_impl.handles_run_events_in_store_event
                and event.is_dagster_event
                and event.get_dagster_event().is_job_event
            ):
                self._run_storage_impl.handle_run_event(
                    run_id, event.get_dagster_event(), datetime_from_timestamp(event.timestamp)
                )
                run = self._instance.get_run_by_id(run_id)
                if (
                    run
                    and event.get_dagster_event().is_run_failure
                    and self._instance.run_retries_enabled
                ):
                    # Note that this tag is only applied to runs that fail. Successful runs will not
                    # have a WILL_RETRY_TAG tag.
                    run_failure_reason = (
                        RunFailureReason(run.tags.get(RUN_FAILURE_REASON_TAG))
                        if run.tags.get(RUN_FAILURE_REASON_TAG)
                        else None
                    )
                    self._instance.add_run_tags(
                        run_id,
                        {
                            WILL_RETRY_TAG: str(
                                self._should_retry_run(run, run_failure_reason)
                            ).lower()
                        },
                    )
            for sub in self._subscribers_impl[run_id]:
                sub(event)

    def _should_retry_run(self, run, run_failure_reason):
        """Helper method to check if run should be retried."""
        from typing import cast

        from dagster._daemon.auto_run_reexecution.auto_run_reexecution import (
            auto_reexecution_should_retry_run,
        )

        # Cast is safe since this mixin is only used by DagsterInstance
        return auto_reexecution_should_retry_run(
            cast("DagsterInstance", self), run, run_failure_reason
        )

    def add_event_listener(self, run_id: str, cb) -> None:
        """Add event listener."""
        self._subscribers_impl[run_id].append(cb)

    def report_engine_event(
        self,
        message: str,
        dagster_run: Optional["DagsterRun"] = None,
        engine_event_data: Optional["EngineEventData"] = None,
        cls: type[object] | None = None,
        step_key: str | None = None,
        job_name: str | None = None,
        run_id: str | None = None,
    ) -> "DagsterEvent":
        """Report a EngineEvent that occurred outside of a job execution context."""
        from dagster._core.events import DagsterEvent, DagsterEventType, EngineEventData
        from dagster._core.storage.dagster_run import DagsterRun

        check.opt_class_param(cls, "cls")
        check.str_param(message, "message")
        check.opt_inst_param(dagster_run, "dagster_run", DagsterRun)
        check.opt_str_param(run_id, "run_id")
        check.opt_str_param(job_name, "job_name")

        check.invariant(
            dagster_run or (job_name and run_id),
            "Must include either dagster_run or job_name and run_id",
        )

        run_id = run_id if run_id else dagster_run.run_id  # type: ignore
        job_name = job_name if job_name else dagster_run.job_name  # type: ignore

        engine_event_data = check.opt_inst_param(
            engine_event_data,
            "engine_event_data",
            EngineEventData,
            EngineEventData({}),
        )

        if cls:
            message = f"[{cls.__name__}] {message}"

        log_level = logging.INFO
        if engine_event_data and engine_event_data.error:
            log_level = logging.ERROR

        dagster_event = DagsterEvent(
            event_type_value=DagsterEventType.ENGINE_EVENT.value,
            job_name=job_name,
            message=message,
            event_specific_data=engine_event_data,
            step_key=step_key,
        )
        self.report_dagster_event(dagster_event, run_id=run_id, log_level=log_level)
        return dagster_event

    def report_dagster_event(
        self,
        dagster_event: "DagsterEvent",
        run_id: str,
        log_level: str | int = logging.INFO,
        batch_metadata: Optional["DagsterEventBatchMetadata"] = None,
        timestamp: float | None = None,
    ) -> None:
        """Takes a DagsterEvent and stores it in persistent storage for the corresponding DagsterRun."""
        from dagster._core.events.log import EventLogEntry

        event_record = EventLogEntry(
            user_message="",
            level=log_level,
            job_name=dagster_event.job_name,
            run_id=run_id,
            error_info=None,
            timestamp=timestamp or get_current_timestamp(),
            step_key=dagster_event.step_key,
            dagster_event=dagster_event,
        )
        self.handle_new_event(event_record, batch_metadata=batch_metadata)

    def report_run_canceling(self, run: "DagsterRun", message: str | None = None) -> None:
        """Report run canceling event."""
        from dagster._core.events import DagsterEvent, DagsterEventType
        from dagster._core.storage.dagster_run import DagsterRun

        check.inst_param(run, "run", DagsterRun)
        message = check.opt_str_param(
            message,
            "message",
            "Sending run termination request.",
        )
        canceling_event = DagsterEvent(
            event_type_value=DagsterEventType.PIPELINE_CANCELING.value,
            job_name=run.job_name,
            message=message,
        )
        self.report_dagster_event(canceling_event, run_id=run.run_id)

    def report_run_canceled(
        self,
        dagster_run: "DagsterRun",
        message: str | None = None,
    ) -> "DagsterEvent":
        """Report run canceled event."""
        from dagster._core.events import DagsterEvent, DagsterEventType
        from dagster._core.storage.dagster_run import DagsterRun

        check.inst_param(dagster_run, "dagster_run", DagsterRun)

        message = check.opt_str_param(
            message,
            "mesage",
            "This run has been marked as canceled from outside the execution context.",
        )

        dagster_event = DagsterEvent(
            event_type_value=DagsterEventType.PIPELINE_CANCELED.value,
            job_name=dagster_run.job_name,
            message=message,
        )
        self.report_dagster_event(dagster_event, run_id=dagster_run.run_id, log_level=logging.ERROR)
        return dagster_event

    def report_run_failed(
        self,
        dagster_run: "DagsterRun",
        message: str | None = None,
        job_failure_data: Optional["JobFailureData"] = None,
    ) -> "DagsterEvent":
        """Report run failed event."""
        from dagster._core.events import DagsterEvent, DagsterEventType
        from dagster._core.storage.dagster_run import DagsterRun

        check.inst_param(dagster_run, "dagster_run", DagsterRun)

        message = check.opt_str_param(
            message,
            "message",
            "This run has been marked as failed from outside the execution context.",
        )

        dagster_event = DagsterEvent(
            event_type_value=DagsterEventType.PIPELINE_FAILURE.value,
            job_name=dagster_run.job_name,
            message=message,
            event_specific_data=job_failure_data,
        )
        self.report_dagster_event(dagster_event, run_id=dagster_run.run_id, log_level=logging.ERROR)
        return dagster_event

    def get_yaml_python_handlers(self) -> Sequence[logging.Handler]:
        """Get YAML-defined Python logging handlers."""
        if self._settings_impl:
            logging_config = self._instance.get_python_log_dagster_handler_config()

            if logging_config:
                beta_warning("Handling yaml-defined logging configuration")

            # Handlers can only be retrieved from dictConfig configuration if they are attached
            # to a logger. We add a dummy logger to the configuration that allows us to access user
            # defined handlers.
            handler_names = logging_config.get("handlers", {}).keys()

            dagster_dummy_logger_name = "dagster_dummy_logger"

            processed_dict_conf = {
                "version": 1,
                "disable_existing_loggers": False,
                "loggers": {dagster_dummy_logger_name: {"handlers": handler_names}},
            }
            processed_dict_conf.update(logging_config)

            logging.config.dictConfig(processed_dict_conf)

            dummy_logger = logging.getLogger(dagster_dummy_logger_name)
            return dummy_logger.handlers
        return []

    def get_event_log_handler(self) -> "_EventListenerLogHandler":
        """Get event log handler."""
        from typing import cast

        from dagster._core.instance.types import _EventListenerLogHandler

        # Cast is safe since this mixin is only used by DagsterInstance
        event_log_handler = _EventListenerLogHandler(cast("DagsterInstance", self))
        event_log_handler.setLevel(10)
        return event_log_handler

    def get_handlers(self) -> Sequence[logging.Handler]:
        """Get all logging handlers."""
        handlers: list[logging.Handler] = [self.get_event_log_handler()]
        handlers.extend(self.get_yaml_python_handlers())
        return handlers



import datetime
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from contextvars import ContextVar
from functools import wraps
from typing import TYPE_CHECKING, Annotated, Concatenate, Optional, Protocol, TypeVar

from dagster_shared.record import ImportFrom, replace
from typing_extensions import ParamSpec

import dagster._check as check
from dagster._core.definitions.temporal_context import TemporalContext
from dagster._record import record
from dagster._symbol_annotations.public import public
from dagster._time import get_current_datetime

if TYPE_CHECKING:
    from dagster._core.instance import DynamicPartitionsStore

P = ParamSpec("P")
T_Return = TypeVar("T_Return")


@record
class PartitionLoadingContext:
    """PartitionLoadingContext is a context object that is passed the partition keys functions of a
    PartitionedJobDefinition. It contains information about where partitions are being loaded from
    and the effective time for the partition loading.

    temporal_context (TemporalContext): The TemporalContext for partition loading.
    dynamic_partitions_store: The DynamicPartitionsStore backing the partition loading.  Used for dynamic partitions definitions.
    """

    temporal_context: TemporalContext
    dynamic_partitions_store: (
        Annotated["DynamicPartitionsStore", ImportFrom("dagster._core.instance")] | None
    )

    @property
    def effective_dt(self) -> datetime.datetime:
        return self.temporal_context.effective_dt

    def updated(
        self,
        effective_dt: datetime.datetime | None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"],
    ) -> "PartitionLoadingContext":
        """Returns a new PartitionLoadingContext with the updated effective_dt and dynamic_partitions_store.
        If the effective_dt and dynamic_partitions_store are unset, the existing context is returned.

        Args:
            effective_dt: The effective time for the partition loading.
            dynamic_partitions_store: The DynamicPartitionsStore backing the partition loading.
        """
        if effective_dt is None and dynamic_partitions_store is None:
            return self

        return replace(
            self,
            temporal_context=replace(
                self.temporal_context, effective_dt=effective_dt or self.effective_dt
            ),
            dynamic_partitions_store=dynamic_partitions_store or self.dynamic_partitions_store,
        )


_current_ctx: ContextVar[PartitionLoadingContext | None] = ContextVar(
    "current_partition_loading_context", default=None
)


def require_full_partition_loading_context(func: Callable[P, T_Return]) -> Callable[P, T_Return]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T_Return:
        current_context = _current_ctx.get()
        check.invariant(
            current_context is not None
            and current_context.effective_dt is not None
            and current_context.dynamic_partitions_store is not None,
            "This function can only be called within a partition_loading_context with both a datetime and dynamic_partitions_store set",
        )
        return func(*args, **kwargs)

    return wrapper


@public
@contextmanager
def partition_loading_context(
    effective_dt: datetime.datetime | None = None,
    dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    *,
    new_ctx: PartitionLoadingContext | None = None,
) -> Iterator[PartitionLoadingContext]:
    """Context manager for setting the current PartitionLoadingContext, which controls how PartitionsDefinitions,
    PartitionMappings, and PartitionSubsets are loaded. This contextmanager is additive, meaning if effective_dt
    or dynamic_partitions_store are not provided, the value from the previous PartitionLoadingContext is used if
    it exists.

    Args:
        effective_dt (Optional[datetime.datetime]): The effective time for the partition loading.
        dynamic_partitions_store (Optional[DynamicPartitionsStore]): The DynamicPartitionsStore backing the partition loading.
        new_ctx (Optional[PartitionLoadingContext]): A new PartitionLoadingContext which will override the current one.

    Examples:
        .. code-block:: python

            import dagster as dg
            import datetime

            partitions_def = dg.DailyPartitionsDefinition(start_date="2021-01-01")

            with dg.partition_loading_context(effective_dt=datetime.datetime(2021, 1, 2)):
                assert partitions_def.get_last_partition_key() == "2021-01-01"

            with dg.partition_loading_context(effective_dt=datetime.datetime(2021, 1, 3)):
                assert partitions_def.get_last_partition_key() == "2021-01-02"
    """
    prev_ctx = _current_ctx.get() or PartitionLoadingContext(
        temporal_context=TemporalContext(effective_dt=get_current_datetime(), last_event_id=None),
        dynamic_partitions_store=None,
    )
    new_ctx = new_ctx or PartitionLoadingContext.updated(
        prev_ctx, effective_dt, dynamic_partitions_store
    )

    token = _current_ctx.set(new_ctx)

    try:
        yield new_ctx
    finally:
        _current_ctx.reset(token)


class _HasPartitionLoadingContext(Protocol):
    _partition_loading_context: PartitionLoadingContext


Self = TypeVar("Self", bound=_HasPartitionLoadingContext)


def use_partition_loading_context(
    func: Callable[Concatenate[Self, P], T_Return],
) -> Callable[Concatenate[Self, P], T_Return]:
    """Decorator for methods that will use the partition loading context."""

    @wraps(func)
    def wrapper(self: Self, *args: P.args, **kwargs: P.kwargs) -> T_Return:
        with partition_loading_context(new_ctx=self._partition_loading_context):
            return func(self, *args, **kwargs)

    return wrapper



import os
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from dagster_shared.record import as_dict

if TYPE_CHECKING:
    from dagster._core.storage.dagster_run import DagsterRun

# 'airflow_execution_date' and 'is_airflow_ingest_pipeline' are hardcoded tags used in the
# airflow ingestion logic (see: dagster_pipeline_factory.py). 'airflow_execution_date' stores the
# 'execution_date' used in Airflow operator execution and 'is_airflow_ingest_pipeline' determines
# whether 'airflow_execution_date' is needed.
# https://github.com/dagster-io/dagster/issues/2403
AIRFLOW_EXECUTION_DATE_STR = "airflow_execution_date"
IS_AIRFLOW_INGEST_PIPELINE_STR = "is_airflow_ingest_pipeline"

# Our internal guts can handle empty strings for job name and run id
# However making these named constants for documentation, to encode where we are making the assumption,
# and to allow us to change this more easily in the future, provided we are disciplined about
# actually using this constants.
RUNLESS_RUN_ID = ""
RUNLESS_JOB_NAME = ""


# The DAGSTER_EVENT_BATCH_SIZE env var controls whether we store events in batches. When enabled,
# this causes _dagster_event_batch_metadata to be injected into log records, and we will then use
# the store_event_batch method of EventLogStorage to persist a batch of events at once instead of
# calling store_event individually. This is expected for cloud deployments, but should only have a
# minimal impact for OSS since the batch is sent over a thread queue.
#
# When the batch size is set to 0, we bypass the batching mechanism almost
# entirely (multiple store_event calls are made instead of store_event_batch). This makes batching
# opt-in.
#
# Note that we don't store the value in the constant so that it can be changed without a process
# restart.
def _get_event_batch_size() -> int:
    return int(os.getenv("DAGSTER_EVENT_BATCH_SIZE", "0"))


def _is_batch_writing_enabled() -> bool:
    return _get_event_batch_size() > 0


def _check_run_equality(
    pipeline_run: "DagsterRun", candidate_run: "DagsterRun"
) -> Mapping[str, tuple[Any, Any]]:
    field_diff: dict[str, tuple[Any, Any]] = {}
    for field, expected_value in as_dict(pipeline_run):
        candidate_value = getattr(candidate_run, field)
        if expected_value != candidate_value:
            field_diff[field] = (expected_value, candidate_value)

    return field_diff


def _format_field_diff(field_diff: Mapping[str, tuple[Any, Any]]) -> str:
    return "\n".join(
        [
            (
                "    {field_name}:\n"
                + "        Expected: {expected_value}\n"
                + "        Received: {candidate_value}"
            ).format(
                field_name=field_name,
                expected_value=expected_value,
                candidate_value=candidate_value,
            )
            for field_name, (
                expected_value,
                candidate_value,
            ) in field_diff.items()
        ]
    )



from dagster._core.definitions.partitions.utils.base import (
    generate_partition_key_based_definition_id as generate_partition_key_based_definition_id,
    raise_error_on_duplicate_partition_keys as raise_error_on_duplicate_partition_keys,
    raise_error_on_invalid_partition_key_substring as raise_error_on_invalid_partition_key_substring,
)
from dagster._core.definitions.partitions.utils.mapping import (
    get_builtin_partition_mapping_types as get_builtin_partition_mapping_types,
    infer_partition_mapping as infer_partition_mapping,
    warn_if_partition_mapping_not_builtin as warn_if_partition_mapping_not_builtin,
)
from dagster._core.definitions.partitions.utils.multi import (
    MultiPartitionKey as MultiPartitionKey,
    PartitionDimensionDefinition as PartitionDimensionDefinition,
    get_multipartition_key_from_tags as get_multipartition_key_from_tags,
    get_tags_from_multi_partition_key as get_tags_from_multi_partition_key,
    get_time_partition_key as get_time_partition_key,
    get_time_partitions_def as get_time_partitions_def,
    has_one_dimension_time_window_partitioning as has_one_dimension_time_window_partitioning,
)
from dagster._core.definitions.partitions.utils.time_window import (
    PartitionRangeStatus as PartitionRangeStatus,
    PartitionTimeWindowStatus as PartitionTimeWindowStatus,
    PersistedTimeWindow as PersistedTimeWindow,
    TimeWindow as TimeWindow,
)



from collections.abc import Callable, Mapping, Sequence

import dagster._check as check
from dagster._annotations import deprecated_param, public
from dagster._core.definitions.partitions.definition.static import StaticPartitionsDefinition
from dagster._core.definitions.partitions.partitioned_config import (
    PartitionConfigFn,
    PartitionedConfig,
)
from dagster._utils.warnings import normalize_renamed_param


@public
@deprecated_param(
    param="tags_for_partition_fn",
    breaking_version="2.0",
    additional_warn_text="Use tags_for_partition_key_fn instead.",
)
def static_partitioned_config(
    partition_keys: Sequence[str],
    tags_for_partition_fn: Callable[[str], Mapping[str, str]] | None = None,
    tags_for_partition_key_fn: Callable[[str], Mapping[str, str]] | None = None,
) -> Callable[[PartitionConfigFn], PartitionedConfig[StaticPartitionsDefinition]]:
    """Creates a static partitioned config for a job.

    The provided partition_keys is a static list of strings identifying the set of partitions. The
    list of partitions is static, so while the run config returned by the decorated function may
    change over time, the list of valid partition keys does not.

    This has performance advantages over `dynamic_partitioned_config` in terms of loading different
    partition views in the Dagster UI.

    The decorated function takes in a partition key and returns a valid run config for a particular
    target job.

    Args:
        partition_keys (Sequence[str]): A list of valid partition keys, which serve as the range of
            values that can be provided to the decorated run config function.
        tags_for_partition_fn (Optional[Callable[[str], Mapping[str, str]]]): A function that
            accepts a partition key and returns a dictionary of tags to attach to runs for that
            partition.
        tags_for_partition_key_fn (Optional[Callable[[str], Mapping[str, str]]]): A function that
            accepts a partition key and returns a dictionary of tags to attach to runs for that
            partition.

    Returns:
        PartitionedConfig
    """
    check.sequence_param(partition_keys, "partition_keys", str)

    tags_for_partition_key_fn = normalize_renamed_param(
        tags_for_partition_key_fn,
        "tags_for_partition_key_fn",
        tags_for_partition_fn,
        "tags_for_partition_fn",
    )

    def inner(
        fn: PartitionConfigFn,
    ) -> PartitionedConfig[StaticPartitionsDefinition]:
        return PartitionedConfig(
            partitions_def=StaticPartitionsDefinition(partition_keys),
            run_config_for_partition_key_fn=fn,
            decorated_fn=fn,
            tags_for_partition_key_fn=tags_for_partition_key_fn,
        )

    return inner



from dagster.components.component.component import (
    Component as Component,
    ComponentTypeSpec as ComponentTypeSpec,
)
from dagster.components.component.component_loader import component_instance as component_instance
from dagster.components.component.state_backed_component import (
    StateBackedComponent as StateBackedComponent,
)
from dagster.components.component_scaffolding import scaffold_component as scaffold_component
from dagster.components.components import (
    DefinitionsComponent as DefinitionsComponent,  # back-compat
    DefsFolderComponent as DefsFolderComponent,
)
from dagster.components.core.context import ComponentLoadContext as ComponentLoadContext
from dagster.components.core.load_defs import (
    build_component_defs as build_component_defs,
    build_defs_for_component as build_defs_for_component,
    load_defs as load_defs,
    load_from_defs_folder as load_from_defs_folder,
)
from dagster.components.definitions import definitions as definitions
from dagster.components.lib.executable_component.function_component import (
    FunctionComponent as FunctionComponent,
)
from dagster.components.lib.executable_component.python_script_component import (
    PythonScriptComponent as PythonScriptComponent,
)
from dagster.components.lib.executable_component.uv_run_component import (
    UvRunComponent as UvRunComponent,
)
from dagster.components.lib.sql_component.sql_component import (
    SqlComponent as SqlComponent,
    TemplatedSqlComponent as TemplatedSqlComponent,
)
from dagster.components.resolved.base import Resolvable as Resolvable
from dagster.components.resolved.context import ResolutionContext as ResolutionContext
from dagster.components.resolved.core_models import (
    AssetAttributesModel as AssetAttributesModel,
    ResolvedAssetCheckSpec as ResolvedAssetCheckSpec,
    ResolvedAssetKey as ResolvedAssetKey,
    ResolvedAssetSpec as ResolvedAssetSpec,
)
from dagster.components.resolved.model import (
    Injected as Injected,
    Model as Model,
    Resolver as Resolver,
)
from dagster.components.scaffold.scaffold import (
    Scaffolder as Scaffolder,
    ScaffoldRequest as ScaffoldRequest,
    scaffold_with as scaffold_with,
)
from dagster.components.utils.defs_state import (
    DefsStateConfig as DefsStateConfig,
    DefsStateConfigArgs as DefsStateConfigArgs,
    ResolvedDefsStateConfig as ResolvedDefsStateConfig,
)



"""Evaluators for each freshness policy type."""

from abc import ABC, abstractmethod

import structlog

import dagster._check as check
from dagster._core.definitions.assets.graph.base_asset_graph import BaseAssetNode
from dagster._core.definitions.freshness import (
    CronFreshnessPolicy,
    FreshnessPolicy,
    FreshnessState,
    TimeWindowFreshnessPolicy,
)
from dagster._core.storage.event_log.base import AssetRecord
from dagster._core.workspace.context import LoadingContext
from dagster._time import get_current_datetime, get_current_timestamp
from dagster._utils.schedules import get_latest_completed_cron_tick

logger = structlog.get_logger(__name__)


# IMPROVEME [OPER-1796] move to oss
class FreshnessPolicyEvaluator(ABC):
    """Abstract base class for freshness policy evaluators.
    Do not implement this class, implement subclasses for each policy type.
    """

    @abstractmethod
    async def evaluate_freshness(
        self, context: LoadingContext, node: BaseAssetNode
    ) -> FreshnessState:
        """Evaluate the freshness of an asset based on the freshness policy, returning the corresponding FreshnessState."""
        raise NotImplementedError("Subclasses must implement this method")


# IMPROVEME [OPER-1796] move to oss
class TimeWindowFreshnessPolicyEvaluator(FreshnessPolicyEvaluator):
    policy_type = TimeWindowFreshnessPolicy

    async def evaluate_freshness(
        self, context: LoadingContext, node: BaseAssetNode
    ) -> FreshnessState:
        """Evaluate the freshness state of an asset based on the time window freshness policy.
        If the asset has no materialization events, return UNKNOWN.
        If the asset was materialized within the policy time window, return PASS.
        If the asset was materialized outside the policy time window, return FAIL.
        If the asset was materialized within the policy time window but outside the warning time window (if provided), return WARN.
        """
        freshness_policy = check.not_none(node.freshness_policy_or_from_metadata)

        time_window_policy = check.inst(
            freshness_policy,
            self.policy_type,
            f"Expected freshness policy type {self.policy_type}, got {freshness_policy.__class__}",
        )

        record = await AssetRecord.gen(context, node.key)
        latest_materialization_event = record.asset_entry.last_materialization if record else None
        if not latest_materialization_event:
            # Freshness cannot be evaluated if there are no materialization events
            return FreshnessState.UNKNOWN

        current_timestamp = get_current_timestamp()
        latest_materialization_timestamp = latest_materialization_event.timestamp

        seconds_since_last_materialization = current_timestamp - latest_materialization_timestamp

        warn_threshold_seconds = (
            time_window_policy.warn_window.to_timedelta().total_seconds()
            if time_window_policy.warn_window
            else None
        )
        fail_threshold_seconds = time_window_policy.fail_window.to_timedelta().total_seconds()

        if seconds_since_last_materialization >= fail_threshold_seconds:
            return FreshnessState.FAIL
        elif (
            warn_threshold_seconds and seconds_since_last_materialization >= warn_threshold_seconds
        ):
            return FreshnessState.WARN
        else:
            return FreshnessState.PASS


class CronFreshnessPolicyEvaluator(FreshnessPolicyEvaluator):
    policy_type = CronFreshnessPolicy

    async def evaluate_freshness(
        self, context: LoadingContext, node: BaseAssetNode
    ) -> FreshnessState:
        """Always look at the time window around the last completed cron tick.
        Time window = [last_completed_cron_tick - lower_bound_delta, last_completed_cron_tick].

        This means that if the current time is within a cron window, we still look at the PREVIOUS window.
        Ex: deadline_cron = "0 9 * * *", lower_bound_delta = 1 hour - asset is expected to materialize between 8-9am every day
        If the current time is 8:59am, we look at YESTERDAY's 8-9am window

        If there is a materialization AFTER the lower bound of the window, asset is fresh, otherwise it is stale.

        Examples:
        asset is expected to materialize beteween 8:45 - 9am every day
        deadline_cron = "0 9 * * *"
        lower_bound_delta = 15 minutes

        Example 1: Asset is on time
        Materialization time: 8:50am
        Asset is fresh starting at 9am, and will remain fresh until at least 9am next day

        Example 2: Asset is late
        Materialization time: 9:30am
        Asset is stale starting at 9am, becomes fresh at 9:30am, will remain fresh until at least 9am next day

        Example 3: Asset is early
        Materialization time: 8:30am
        Asset is stale starting at 9am, will remain stale until it materializes again.
        """
        # Get the policy, making sure it's a cron policy
        freshness_policy = check.not_none(node.freshness_policy_or_from_metadata)

        cron_policy = check.inst(
            freshness_policy,
            self.policy_type,
            f"Expected freshness policy type {self.policy_type}, got {freshness_policy.__class__}",
        )

        # Get the latest materialization event
        record = await AssetRecord.gen(context, node.key)
        latest_materialization_event = record.asset_entry.last_materialization if record else None
        # We can't evaluate freshness if there are no materialization events
        if not latest_materialization_event:
            return FreshnessState.UNKNOWN

        # Get the current datetime, in the timezone set on the policy
        current_datetime = get_current_datetime(cron_policy.timezone)

        # Get the timestamp of the last completed cron tick, evaluated in the timezone set on the policy
        latest_completed_cron_tick = get_latest_completed_cron_tick(
            cron_policy.deadline_cron, current_datetime, cron_policy.timezone
        )

        # Get the start of the last completed cron window, evaluated in the timezone set on the policy
        # Seconds resolution is okay here since we don't allow higher than to-the-minute resolution in the deadline cron
        cron_window_start = (
            latest_completed_cron_tick.timestamp() - cron_policy.lower_bound_delta.total_seconds()
        )

        # If the latest materialization is after the lower bound of the window, the asset is fresh, otherwise it is stale.
        return (
            FreshnessState.PASS
            if latest_materialization_event.timestamp >= cron_window_start
            else FreshnessState.FAIL
        )


FRESHNESS_EVALUATORS_BY_POLICY_TYPE: dict[type[FreshnessPolicy], type[FreshnessPolicyEvaluator]] = {
    TimeWindowFreshnessPolicy: TimeWindowFreshnessPolicyEvaluator,
    CronFreshnessPolicy: CronFreshnessPolicyEvaluator,
}



# Import and re-export commonly used symbols for backwards compatibility
from collections.abc import Mapping
from typing import Any, TypeAlias

from dagster._core.instance.instance import DagsterInstance as DagsterInstance
from dagster._core.instance.ref import InstanceRef as InstanceRef
from dagster._core.instance.types import (
    CachingDynamicPartitionsLoader as CachingDynamicPartitionsLoader,
    DynamicPartitionsStore as DynamicPartitionsStore,
    InstanceType as InstanceType,
    MayHaveInstanceWeakref as MayHaveInstanceWeakref,
    T_DagsterInstance as T_DagsterInstance,
)
from dagster._core.instance.utils import (
    RUNLESS_JOB_NAME as RUNLESS_JOB_NAME,
    RUNLESS_RUN_ID as RUNLESS_RUN_ID,
)

# Re-export for backward compatibility - internal repo depends on this export from instance module
from dagster._time import get_current_timestamp as get_current_timestamp

# Type alias for backward compatibility
DagsterInstanceOverrides: TypeAlias = Mapping[str, Any]

# Re-export for backward compatibility
__all__ = [
    "RUNLESS_JOB_NAME",
    "RUNLESS_RUN_ID",
    "DagsterInstance",
    "DagsterInstanceOverrides",
    "DynamicPartitionsStore",
    "InstanceRef",
    "InstanceType",
    "MayHaveInstanceWeakref",
    "T_DagsterInstance",
    "get_current_timestamp",
]
